/*
 * Copyright (c) 2017 Digital Bazaar, Inc. All rights reserved.
 */
'use strict';

function _toArray(arr) { return Array.isArray(arr) ? arr : Array.from(arr); }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var JsonLdError = require('./JsonLdError');

var _require = require('./types'),
    _isArray = _require.isArray,
    _isObject = _require.isObject,
    _isString = _require.isString,
    _isUndefined = _require.isUndefined;

var _require2 = require('./graphTypes'),
    _isList = _require2.isList,
    _isValue = _require2.isValue,
    _isGraph = _require2.isGraph,
    _isSimpleGraph = _require2.isSimpleGraph,
    _isSubjectReference = _require2.isSubjectReference;

var _require3 = require('./context'),
    _expandIri = _require3.expandIri,
    _getContextValue = _require3.getContextValue,
    _isKeyword = _require3.isKeyword,
    _processContext = _require3.process;

var _require4 = require('./url'),
    _removeBase = _require4.removeBase;

var _require5 = require('./util'),
    _addValue = _require5.addValue,
    _asArray = _require5.asArray,
    _compareShortestLeast = _require5.compareShortestLeast;

var api = {};
module.exports = api;

/**
 * Recursively compacts an element using the given active context. All values
 * must be in expanded form before this method is called.
 *
 * @param activeCtx the active context to use.
 * @param activeProperty the compacted property associated with the element
 *          to compact, null for none.
 * @param element the element to compact.
 * @param options the compaction options.
 * @param compactionMap the compaction map to use.
 *
 * @return the compacted value.
 */
api.compact = function (_ref) {
  var activeCtx = _ref.activeCtx,
      _ref$activeProperty = _ref.activeProperty,
      activeProperty = _ref$activeProperty === undefined ? null : _ref$activeProperty,
      element = _ref.element,
      _ref$options = _ref.options,
      options = _ref$options === undefined ? {} : _ref$options,
      _ref$compactionMap = _ref.compactionMap,
      compactionMap = _ref$compactionMap === undefined ? function () {
    return undefined;
  } : _ref$compactionMap;

  // recursively compact array
  if (_isArray(element)) {
    var rval = [];
    for (var i = 0; i < element.length; ++i) {
      // compact, dropping any null values unless custom mapped
      var compacted = api.compact({
        activeCtx: activeCtx,
        activeProperty: activeProperty,
        element: element[i],
        options: options,
        compactionMap: compactionMap
      });
      if (compacted === null) {
        // TODO: use `await` to support async
        compacted = compactionMap({
          unmappedValue: element[i],
          activeCtx: activeCtx,
          activeProperty: activeProperty,
          parent: element,
          index: i,
          options: options
        });
        if (compacted === undefined) {
          continue;
        }
      }
      rval.push(compacted);
    }
    if (options.compactArrays && rval.length === 1) {
      // use single element if no container is specified
      var container = _getContextValue(activeCtx, activeProperty, '@container') || [];
      if (container.length === 0) {
        rval = rval[0];
      }
    }
    return rval;
  }

  // use any scoped context on activeProperty
  var ctx = _getContextValue(activeCtx, activeProperty, '@context');
  if (!_isUndefined(ctx)) {
    // Note: spec's `from term` var is named `isPropertyTermScopedContext`
    activeCtx = _processContext({
      activeCtx: activeCtx,
      localCtx: ctx,
      isPropertyTermScopedContext: true,
      options: options
    });
  }

  // recursively compact object
  if (_isObject(element)) {
    if (options.link && '@id' in element && options.link.hasOwnProperty(element['@id'])) {
      // check for a linked element to reuse
      var linked = options.link[element['@id']];
      for (var _i = 0; _i < linked.length; ++_i) {
        if (linked[_i].expanded === element) {
          return linked[_i].compacted;
        }
      }
    }

    // do value compaction on @values and subject references
    if (_isValue(element) || _isSubjectReference(element)) {
      var _rval2 = api.compactValue({ activeCtx: activeCtx, activeProperty: activeProperty, value: element, options: options });
      if (options.link && _isSubjectReference(element)) {
        // store linked element
        if (!options.link.hasOwnProperty(element['@id'])) {
          options.link[element['@id']] = [];
        }
        options.link[element['@id']].push({ expanded: element, compacted: _rval2 });
      }
      return _rval2;
    }

    // FIXME: avoid misuse of active property as an expanded property?
    var insideReverse = activeProperty === '@reverse';

    var _rval = {};

    // revert type scoped context
    activeCtx = activeCtx.revertTypeScopedContext();

    if (options.link && '@id' in element) {
      // store linked element
      if (!options.link.hasOwnProperty(element['@id'])) {
        options.link[element['@id']] = [];
      }
      options.link[element['@id']].push({ expanded: element, compacted: _rval });
    }

    // apply any context defined on an alias of @type
    // if key is @type and any compacted value is a term having a local
    // context, overlay that context
    var types = element['@type'] || [];
    if (types.length > 1) {
      types = Array.from(types).sort();
    }
    // find all type-scoped contexts based on current context, prior to
    // updating it
    var typeContext = activeCtx;
    var _iteratorNormalCompletion = true;
    var _didIteratorError = false;
    var _iteratorError = undefined;

    try {
      for (var _iterator = types[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
        var type = _step.value;

        var compactedType = api.compactIri({ activeCtx: typeContext, iri: type, relativeTo: { vocab: true } });

        // Use any type-scoped context defined on this value
        var _ctx = _getContextValue(typeContext, compactedType, '@context');
        if (!_isUndefined(_ctx)) {
          activeCtx = _processContext({
            activeCtx: activeCtx,
            localCtx: _ctx,
            options: options,
            isTypeScopedContext: true
          });
        }
      }

      // process element keys in order
    } catch (err) {
      _didIteratorError = true;
      _iteratorError = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion && _iterator.return) {
          _iterator.return();
        }
      } finally {
        if (_didIteratorError) {
          throw _iteratorError;
        }
      }
    }

    var keys = Object.keys(element).sort();
    var _iteratorNormalCompletion2 = true;
    var _didIteratorError2 = false;
    var _iteratorError2 = undefined;

    try {
      for (var _iterator2 = keys[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
        var expandedProperty = _step2.value;

        var expandedValue = element[expandedProperty];

        // compact @id and @type(s)
        if (expandedProperty === '@id' || expandedProperty === '@type') {
          var _ret = function () {
            // if using a type-scoped context, resolve type values against previous
            // context
            var isType = expandedProperty === '@type';
            var valueContext = isType ? activeCtx.previousContext || activeCtx : activeCtx;
            var compactedValue = _asArray(expandedValue).map(function (expandedIri) {
              return api.compactIri({
                activeCtx: valueContext,
                iri: expandedIri,
                relativeTo: { vocab: isType }
              });
            });
            if (compactedValue.length === 1) {
              compactedValue = compactedValue[0];
            }

            // use keyword alias and add value
            var alias = api.compactIri({ activeCtx: activeCtx, iri: expandedProperty, relativeTo: { vocab: true } });
            var isArray = _isArray(compactedValue) && expandedValue.length === 0;
            _addValue(_rval, alias, compactedValue, { propertyIsArray: isArray });
            return 'continue';
          }();

          if (_ret === 'continue') continue;
        }

        // handle @reverse
        if (expandedProperty === '@reverse') {
          // recursively compact expanded value
          var compactedValue = api.compact({
            activeCtx: activeCtx,
            activeProperty: '@reverse',
            element: expandedValue,
            options: options,
            compactionMap: compactionMap
          });

          // handle double-reversed properties
          for (var compactedProperty in compactedValue) {
            if (activeCtx.mappings.has(compactedProperty) && activeCtx.mappings.get(compactedProperty).reverse) {
              var value = compactedValue[compactedProperty];
              var _container = _getContextValue(activeCtx, compactedProperty, '@container') || [];
              var useArray = _container.includes('@set') || !options.compactArrays;
              _addValue(_rval, compactedProperty, value, { propertyIsArray: useArray });
              delete compactedValue[compactedProperty];
            }
          }

          if (Object.keys(compactedValue).length > 0) {
            // use keyword alias and add value
            var alias = api.compactIri({
              activeCtx: activeCtx,
              iri: expandedProperty,
              relativeTo: { vocab: true }
            });
            _addValue(_rval, alias, compactedValue);
          }

          continue;
        }

        if (expandedProperty === '@preserve') {
          // compact using activeProperty
          var _compactedValue = api.compact({
            activeCtx: activeCtx,
            activeProperty: activeProperty,
            element: expandedValue,
            options: options,
            compactionMap: compactionMap });

          if (!(_isArray(_compactedValue) && _compactedValue.length === 0)) {
            _addValue(_rval, expandedProperty, _compactedValue);
          }
          continue;
        }

        // handle @index property
        if (expandedProperty === '@index') {
          // drop @index if inside an @index container
          var _container2 = _getContextValue(activeCtx, activeProperty, '@container') || [];
          if (_container2.includes('@index')) {
            continue;
          }

          // use keyword alias and add value
          var _alias = api.compactIri({
            activeCtx: activeCtx,
            iri: expandedProperty,
            relativeTo: { vocab: true }
          });
          _addValue(_rval, _alias, expandedValue);
          continue;
        }

        // skip array processing for keywords that aren't @graph or @list
        if (expandedProperty !== '@graph' && expandedProperty !== '@list' && _isKeyword(expandedProperty)) {
          // use keyword alias and add value as is
          var _alias2 = api.compactIri({
            activeCtx: activeCtx,
            iri: expandedProperty,
            relativeTo: { vocab: true }
          });
          _addValue(_rval, _alias2, expandedValue);
          continue;
        }

        // Note: expanded value must be an array due to expansion algorithm.
        if (!_isArray(expandedValue)) {
          throw new JsonLdError('JSON-LD expansion error; expanded value must be an array.', 'jsonld.SyntaxError');
        }

        // preserve empty arrays
        if (expandedValue.length === 0) {
          var itemActiveProperty = api.compactIri({
            activeCtx: activeCtx,
            iri: expandedProperty,
            value: expandedValue,
            relativeTo: { vocab: true },
            reverse: insideReverse
          });
          var nestProperty = activeCtx.mappings.has(itemActiveProperty) ? activeCtx.mappings.get(itemActiveProperty)['@nest'] : null;
          var nestResult = _rval;
          if (nestProperty) {
            _checkNestProperty(activeCtx, nestProperty, options);
            if (!_isObject(_rval[nestProperty])) {
              _rval[nestProperty] = {};
            }
            nestResult = _rval[nestProperty];
          }
          _addValue(nestResult, itemActiveProperty, expandedValue, {
            propertyIsArray: true
          });
        }

        // recusively process array values
        var _iteratorNormalCompletion3 = true;
        var _didIteratorError3 = false;
        var _iteratorError3 = undefined;

        try {
          for (var _iterator3 = expandedValue[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
            var expandedItem = _step3.value;

            // compact property and get container type
            var _itemActiveProperty = api.compactIri({
              activeCtx: activeCtx,
              iri: expandedProperty,
              value: expandedItem,
              relativeTo: { vocab: true },
              reverse: insideReverse
            });

            // if itemActiveProperty is a @nest property, add values to nestResult,
            // otherwise rval
            var _nestProperty = activeCtx.mappings.has(_itemActiveProperty) ? activeCtx.mappings.get(_itemActiveProperty)['@nest'] : null;
            var _nestResult = _rval;
            if (_nestProperty) {
              _checkNestProperty(activeCtx, _nestProperty, options);
              if (!_isObject(_rval[_nestProperty])) {
                _rval[_nestProperty] = {};
              }
              _nestResult = _rval[_nestProperty];
            }

            var _container3 = _getContextValue(activeCtx, _itemActiveProperty, '@container') || [];

            // get simple @graph or @list value if appropriate
            var isGraph = _isGraph(expandedItem);
            var isList = _isList(expandedItem);
            var inner = void 0;
            if (isList) {
              inner = expandedItem['@list'];
            } else if (isGraph) {
              inner = expandedItem['@graph'];
            }

            // recursively compact expanded item
            var compactedItem = api.compact({
              activeCtx: activeCtx,
              activeProperty: _itemActiveProperty,
              element: isList || isGraph ? inner : expandedItem,
              options: options,
              compactionMap: compactionMap
            });

            // handle @list
            if (isList) {
              // ensure @list value is an array
              if (!_isArray(compactedItem)) {
                compactedItem = [compactedItem];
              }

              if (!_container3.includes('@list')) {
                // wrap using @list alias
                compactedItem = _defineProperty({}, api.compactIri({
                  activeCtx: activeCtx,
                  iri: '@list',
                  relativeTo: { vocab: true }
                }), compactedItem);

                // include @index from expanded @list, if any
                if ('@index' in expandedItem) {
                  compactedItem[api.compactIri({
                    activeCtx: activeCtx,
                    iri: '@index',
                    relativeTo: { vocab: true }
                  })] = expandedItem['@index'];
                }
              } else if (_nestResult.hasOwnProperty(_itemActiveProperty)) {
                // can't use @list container for more than 1 list
                throw new JsonLdError('JSON-LD compact error; property has a "@list" @container ' + 'rule but there is more than a single @list that matches ' + 'the compacted term in the document. Compaction might mix ' + 'unwanted items into the list.', 'jsonld.SyntaxError', { code: 'compaction to list of lists' });
              }
            }

            // Graph object compaction cases
            if (isGraph) {
              if (_container3.includes('@graph') && (_container3.includes('@id') || _container3.includes('@index') && _isSimpleGraph(expandedItem))) {
                // get or create the map object
                var mapObject = void 0;
                if (_nestResult.hasOwnProperty(_itemActiveProperty)) {
                  mapObject = _nestResult[_itemActiveProperty];
                } else {
                  _nestResult[_itemActiveProperty] = mapObject = {};
                }

                // index on @id or @index or alias of @none
                var key = (_container3.includes('@id') ? expandedItem['@id'] : expandedItem['@index']) || api.compactIri({ activeCtx: activeCtx, iri: '@none', vocab: true });
                // add compactedItem to map, using value of `@id` or a new blank
                // node identifier

                _addValue(mapObject, key, compactedItem, {
                  propertyIsArray: !options.compactArrays || _container3.includes('@set')
                });
              } else if (_container3.includes('@graph') && _isSimpleGraph(expandedItem)) {
                // container includes @graph but not @id or @index and value is a
                // simple graph object add compact value
                _addValue(_nestResult, _itemActiveProperty, compactedItem, {
                  propertyIsArray: !options.compactArrays || _container3.includes('@set')
                });
              } else {
                // wrap using @graph alias, remove array if only one item and
                // compactArrays not set
                if (_isArray(compactedItem) && compactedItem.length === 1 && options.compactArrays) {
                  compactedItem = compactedItem[0];
                }
                compactedItem = _defineProperty({}, api.compactIri({
                  activeCtx: activeCtx,
                  iri: '@graph',
                  relativeTo: { vocab: true }
                }), compactedItem);

                // include @id from expanded graph, if any
                if ('@id' in expandedItem) {
                  compactedItem[api.compactIri({
                    activeCtx: activeCtx,
                    iri: '@id',
                    relativeTo: { vocab: true }
                  })] = expandedItem['@id'];
                }

                // include @index from expanded graph, if any
                if ('@index' in expandedItem) {
                  compactedItem[api.compactIri({
                    activeCtx: activeCtx,
                    iri: '@index',
                    relativeTo: { vocab: true }
                  })] = expandedItem['@index'];
                }
                _addValue(_nestResult, _itemActiveProperty, compactedItem, {
                  propertyIsArray: !options.compactArrays || _container3.includes('@set')
                });
              }
            } else if (_container3.includes('@language') || _container3.includes('@index') || _container3.includes('@id') || _container3.includes('@type')) {
              // handle language and index maps
              // get or create the map object
              var _mapObject = void 0;
              if (_nestResult.hasOwnProperty(_itemActiveProperty)) {
                _mapObject = _nestResult[_itemActiveProperty];
              } else {
                _nestResult[_itemActiveProperty] = _mapObject = {};
              }

              var _key = void 0;
              if (_container3.includes('@language')) {
                // if container is a language map, simplify compacted value to
                // a simple string
                if (_isValue(compactedItem)) {
                  compactedItem = compactedItem['@value'];
                }
                _key = expandedItem['@language'];
              } else if (_container3.includes('@index')) {
                _key = expandedItem['@index'];
              } else if (_container3.includes('@id')) {
                var idKey = api.compactIri({ activeCtx: activeCtx, iri: '@id', vocab: true });
                _key = compactedItem[idKey];
                delete compactedItem[idKey];
              } else if (_container3.includes('@type')) {
                var typeKey = api.compactIri({
                  activeCtx: activeCtx,
                  iri: '@type',
                  vocab: true
                });
                var _types = void 0;

                var _asArray2 = _asArray(compactedItem[typeKey] || []);

                var _asArray3 = _toArray(_asArray2);

                _key = _asArray3[0];
                _types = _asArray3.slice(1);

                switch (_types.length) {
                  case 0:
                    delete compactedItem[typeKey];
                    break;
                  case 1:
                    compactedItem[typeKey] = _types[0];
                    break;
                  default:
                    compactedItem[typeKey] = _types;
                    break;
                }
              }

              // if compacting this value which has no key, index on @none
              if (!_key) {
                _key = api.compactIri({ activeCtx: activeCtx, iri: '@none', vocab: true });
              }
              // add compact value to map object using key from expanded value
              // based on the container type
              _addValue(_mapObject, _key, compactedItem, {
                propertyIsArray: _container3.includes('@set')
              });
            } else {
              // use an array if: compactArrays flag is false,
              // @container is @set or @list , value is an empty
              // array, or key is @graph
              var isArray = !options.compactArrays || _container3.includes('@set') || _container3.includes('@list') || _isArray(compactedItem) && compactedItem.length === 0 || expandedProperty === '@list' || expandedProperty === '@graph';

              // add compact value
              _addValue(_nestResult, _itemActiveProperty, compactedItem, { propertyIsArray: isArray });
            }
          }
        } catch (err) {
          _didIteratorError3 = true;
          _iteratorError3 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion3 && _iterator3.return) {
              _iterator3.return();
            }
          } finally {
            if (_didIteratorError3) {
              throw _iteratorError3;
            }
          }
        }
      }
    } catch (err) {
      _didIteratorError2 = true;
      _iteratorError2 = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion2 && _iterator2.return) {
          _iterator2.return();
        }
      } finally {
        if (_didIteratorError2) {
          throw _iteratorError2;
        }
      }
    }

    return _rval;
  }

  // only primitives remain which are already compact
  return element;
};

/**
 * Compacts an IRI or keyword into a term or prefix if it can be. If the
 * IRI has an associated value it may be passed.
 *
 * @param activeCtx the active context to use.
 * @param iri the IRI to compact.
 * @param value the value to check or null.
 * @param relativeTo options for how to compact IRIs:
 *          vocab: true to split after @vocab, false not to.
 * @param reverse true if a reverse property is being compacted, false if not.
 *
 * @return the compacted term, prefix, keyword alias, or the original IRI.
 */
api.compactIri = function (_ref2) {
  var activeCtx = _ref2.activeCtx,
      iri = _ref2.iri,
      _ref2$value = _ref2.value,
      value = _ref2$value === undefined ? null : _ref2$value,
      _ref2$relativeTo = _ref2.relativeTo,
      relativeTo = _ref2$relativeTo === undefined ? { vocab: false } : _ref2$relativeTo,
      _ref2$reverse = _ref2.reverse,
      reverse = _ref2$reverse === undefined ? false : _ref2$reverse;

  // can't compact null
  if (iri === null) {
    return iri;
  }

  // if context is from a property term scoped context composed with a
  // type-scoped context, then use the previous context instead
  if (activeCtx.isPropertyTermScoped && activeCtx.previousContext) {
    activeCtx = activeCtx.previousContext;
  }

  var inverseCtx = activeCtx.getInverse();

  // if term is a keyword, it may be compacted to a simple alias
  if (_isKeyword(iri) && iri in inverseCtx && '@none' in inverseCtx[iri] && '@type' in inverseCtx[iri]['@none'] && '@none' in inverseCtx[iri]['@none']['@type']) {
    return inverseCtx[iri]['@none']['@type']['@none'];
  }

  // use inverse context to pick a term if iri is relative to vocab
  if (relativeTo.vocab && iri in inverseCtx) {
    var defaultLanguage = activeCtx['@language'] || '@none';

    // prefer @index if available in value
    var containers = [];
    if (_isObject(value) && '@index' in value && !('@graph' in value)) {
      containers.push('@index', '@index@set');
    }

    // if value is a preserve object, use its value
    if (_isObject(value) && '@preserve' in value) {
      value = value['@preserve'][0];
    }

    // prefer most specific container including @graph, prefering @set
    // variations
    if (_isGraph(value)) {
      // favor indexmap if the graph is indexed
      if ('@index' in value) {
        containers.push('@graph@index', '@graph@index@set', '@index', '@index@set');
      }
      // favor idmap if the graph is has an @id
      if ('@id' in value) {
        containers.push('@graph@id', '@graph@id@set');
      }
      containers.push('@graph', '@graph@set', '@set');
      // allow indexmap if the graph is not indexed
      if (!('@index' in value)) {
        containers.push('@graph@index', '@graph@index@set', '@index', '@index@set');
      }
      // allow idmap if the graph does not have an @id
      if (!('@id' in value)) {
        containers.push('@graph@id', '@graph@id@set');
      }
    } else if (_isObject(value) && !_isValue(value)) {
      containers.push('@id', '@id@set', '@type', '@set@type');
    }

    // defaults for term selection based on type/language
    var typeOrLanguage = '@language';
    var typeOrLanguageValue = '@null';

    if (reverse) {
      typeOrLanguage = '@type';
      typeOrLanguageValue = '@reverse';
      containers.push('@set');
    } else if (_isList(value)) {
      // choose the most specific term that works for all elements in @list
      // only select @list containers if @index is NOT in value
      if (!('@index' in value)) {
        containers.push('@list');
      }
      var list = value['@list'];
      if (list.length === 0) {
        // any empty list can be matched against any term that uses the
        // @list container regardless of @type or @language
        typeOrLanguage = '@any';
        typeOrLanguageValue = '@none';
      } else {
        var commonLanguage = list.length === 0 ? defaultLanguage : null;
        var commonType = null;
        for (var i = 0; i < list.length; ++i) {
          var item = list[i];
          var itemLanguage = '@none';
          var itemType = '@none';
          if (_isValue(item)) {
            if ('@language' in item) {
              itemLanguage = item['@language'];
            } else if ('@type' in item) {
              itemType = item['@type'];
            } else {
              // plain literal
              itemLanguage = '@null';
            }
          } else {
            itemType = '@id';
          }
          if (commonLanguage === null) {
            commonLanguage = itemLanguage;
          } else if (itemLanguage !== commonLanguage && _isValue(item)) {
            commonLanguage = '@none';
          }
          if (commonType === null) {
            commonType = itemType;
          } else if (itemType !== commonType) {
            commonType = '@none';
          }
          // there are different languages and types in the list, so choose
          // the most generic term, no need to keep iterating the list
          if (commonLanguage === '@none' && commonType === '@none') {
            break;
          }
        }
        commonLanguage = commonLanguage || '@none';
        commonType = commonType || '@none';
        if (commonType !== '@none') {
          typeOrLanguage = '@type';
          typeOrLanguageValue = commonType;
        } else {
          typeOrLanguageValue = commonLanguage;
        }
      }
    } else {
      if (_isValue(value)) {
        if ('@language' in value && !('@index' in value)) {
          containers.push('@language', '@language@set');
          typeOrLanguageValue = value['@language'];
        } else if ('@type' in value) {
          typeOrLanguage = '@type';
          typeOrLanguageValue = value['@type'];
        }
      } else {
        typeOrLanguage = '@type';
        typeOrLanguageValue = '@id';
      }
      containers.push('@set');
    }

    // do term selection
    containers.push('@none');

    // an index map can be used to index values using @none, so add as a low
    // priority
    if (_isObject(value) && !('@index' in value)) {
      // allow indexing even if no @index present
      containers.push('@index', '@index@set');
    }

    // values without type or language can use @language map
    if (_isValue(value) && Object.keys(value).length === 1) {
      // allow indexing even if no @index present
      containers.push('@language', '@language@set');
    }

    var term = _selectTerm(activeCtx, iri, value, containers, typeOrLanguage, typeOrLanguageValue);
    if (term !== null) {
      return term;
    }
  }

  // no term match, use @vocab if available
  if (relativeTo.vocab) {
    if ('@vocab' in activeCtx) {
      // determine if vocab is a prefix of the iri
      var vocab = activeCtx['@vocab'];
      if (iri.indexOf(vocab) === 0 && iri !== vocab) {
        // use suffix as relative iri if it is not a term in the active context
        var suffix = iri.substr(vocab.length);
        if (!activeCtx.mappings.has(suffix)) {
          return suffix;
        }
      }
    }
  }

  // no term or @vocab match, check for possible CURIEs
  var choice = null;
  // TODO: make FastCurieMap a class with a method to do this lookup
  var partialMatches = [];
  var iriMap = activeCtx.fastCurieMap;
  // check for partial matches of against `iri`, which means look until
  // iri.length - 1, not full length
  var maxPartialLength = iri.length - 1;
  for (var _i2 = 0; _i2 < maxPartialLength && iri[_i2] in iriMap; ++_i2) {
    iriMap = iriMap[iri[_i2]];
    if ('' in iriMap) {
      partialMatches.push(iriMap[''][0]);
    }
  }
  // check partial matches in reverse order to prefer longest ones first
  for (var _i3 = partialMatches.length - 1; _i3 >= 0; --_i3) {
    var entry = partialMatches[_i3];
    var terms = entry.terms;
    var _iteratorNormalCompletion4 = true;
    var _didIteratorError4 = false;
    var _iteratorError4 = undefined;

    try {
      for (var _iterator4 = terms[Symbol.iterator](), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
        var _term = _step4.value;

        // a CURIE is usable if:
        // 1. it has no mapping, OR
        // 2. value is null, which means we're not compacting an @value, AND
        //   the mapping matches the IRI
        var curie = _term + ':' + iri.substr(entry.iri.length);
        var isUsableCurie = activeCtx.mappings.get(_term)._prefix && (!activeCtx.mappings.has(curie) || value === null && activeCtx.mappings.get(curie)['@id'] === iri);

        // select curie if it is shorter or the same length but lexicographically
        // less than the current choice
        if (isUsableCurie && (choice === null || _compareShortestLeast(curie, choice) < 0)) {
          choice = curie;
        }
      }
    } catch (err) {
      _didIteratorError4 = true;
      _iteratorError4 = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion4 && _iterator4.return) {
          _iterator4.return();
        }
      } finally {
        if (_didIteratorError4) {
          throw _iteratorError4;
        }
      }
    }
  }

  // return chosen curie
  if (choice !== null) {
    return choice;
  }

  // compact IRI relative to base
  if (!relativeTo.vocab) {
    return _removeBase(activeCtx['@base'], iri);
  }

  // return IRI as is
  return iri;
};

/**
 * Performs value compaction on an object with '@value' or '@id' as the only
 * property.
 *
 * @param activeCtx the active context.
 * @param activeProperty the active property that points to the value.
 * @param value the value to compact.
 * @param {Object} [options] - processing options.
 *
 * @return the compaction result.
 */
api.compactValue = function (_ref3) {
  var activeCtx = _ref3.activeCtx,
      activeProperty = _ref3.activeProperty,
      value = _ref3.value,
      options = _ref3.options;

  // value is a @value
  if (_isValue(value)) {
    // get context rules
    var _type = _getContextValue(activeCtx, activeProperty, '@type');
    var language = _getContextValue(activeCtx, activeProperty, '@language');
    var container = _getContextValue(activeCtx, activeProperty, '@container') || [];

    // whether or not the value has an @index that must be preserved
    var preserveIndex = '@index' in value && !container.includes('@index');

    // if there's no @index to preserve ...
    if (!preserveIndex) {
      // matching @type or @language specified in context, compact value
      if (value['@type'] === _type || value['@language'] === language) {
        return value['@value'];
      }
    }

    // return just the value of @value if all are true:
    // 1. @value is the only key or @index isn't being preserved
    // 2. there is no default language or @value is not a string or
    //   the key has a mapping with a null @language
    var keyCount = Object.keys(value).length;
    var isValueOnlyKey = keyCount === 1 || keyCount === 2 && '@index' in value && !preserveIndex;
    var hasDefaultLanguage = '@language' in activeCtx;
    var isValueString = _isString(value['@value']);
    var hasNullMapping = activeCtx.mappings.has(activeProperty) && activeCtx.mappings.get(activeProperty)['@language'] === null;
    if (isValueOnlyKey && (!hasDefaultLanguage || !isValueString || hasNullMapping)) {
      return value['@value'];
    }

    var rval = {};

    // preserve @index
    if (preserveIndex) {
      rval[api.compactIri({
        activeCtx: activeCtx,
        iri: '@index',
        relativeTo: { vocab: true }
      })] = value['@index'];
    }

    if ('@type' in value) {
      // compact @type IRI
      rval[api.compactIri({
        activeCtx: activeCtx,
        iri: '@type',
        relativeTo: { vocab: true }
      })] = api.compactIri({ activeCtx: activeCtx, iri: value['@type'], relativeTo: { vocab: true } });
    } else if ('@language' in value) {
      // alias @language
      rval[api.compactIri({
        activeCtx: activeCtx,
        iri: '@language',
        relativeTo: { vocab: true }
      })] = value['@language'];
    }

    // alias @value
    rval[api.compactIri({
      activeCtx: activeCtx,
      iri: '@value',
      relativeTo: { vocab: true }
    })] = value['@value'];

    return rval;
  }

  // value is a subject reference
  var expandedProperty = _expandIri(activeCtx, activeProperty, { vocab: true }, options);
  var type = _getContextValue(activeCtx, activeProperty, '@type');
  var compacted = api.compactIri({ activeCtx: activeCtx, iri: value['@id'], relativeTo: { vocab: type === '@vocab' } });

  // compact to scalar
  if (type === '@id' || type === '@vocab' || expandedProperty === '@graph') {
    return compacted;
  }

  return _defineProperty({}, api.compactIri({
    activeCtx: activeCtx,
    iri: '@id',
    relativeTo: { vocab: true }
  }), compacted);
};

/**
 * Removes the @preserve keywords as the last step of the compaction
 * algorithm when it is running on framed output.
 *
 * @param ctx the active context used to compact the input.
 * @param input the framed, compacted output.
 * @param options the compaction options used.
 *
 * @return the resulting output.
 */
api.removePreserve = function (ctx, input, options) {
  // recurse through arrays
  if (_isArray(input)) {
    var output = [];
    for (var i = 0; i < input.length; ++i) {
      var result = api.removePreserve(ctx, input[i], options);
      // drop nulls from arrays
      if (result !== null) {
        output.push(result);
      }
    }
    input = output;
  } else if (_isObject(input)) {
    // remove @preserve
    if ('@preserve' in input) {
      if (input['@preserve'] === '@null') {
        return null;
      }
      return input['@preserve'];
    }

    // skip @values
    if (_isValue(input)) {
      return input;
    }

    // recurse through @lists
    if (_isList(input)) {
      input['@list'] = api.removePreserve(ctx, input['@list'], options);
      return input;
    }

    // handle in-memory linked nodes
    var idAlias = api.compactIri({
      activeCtx: ctx,
      iri: '@id',
      relativeTo: { vocab: true }
    });
    if (input.hasOwnProperty(idAlias)) {
      var id = input[idAlias];
      if (options.link.hasOwnProperty(id)) {
        var idx = options.link[id].indexOf(input);
        if (idx !== -1) {
          // already visited
          return options.link[id][idx];
        }
        // prevent circular visitation
        options.link[id].push(input);
      } else {
        // prevent circular visitation
        options.link[id] = [input];
      }
    }

    // recurse through properties
    var graphAlias = api.compactIri({
      activeCtx: ctx,
      iri: '@graph',
      relativeTo: { vocab: true }
    });
    for (var prop in input) {
      // potentially remove the id, if it is an unreference bnode
      if (prop === idAlias && options.bnodesToClear.includes(input[prop])) {
        delete input[idAlias];
        continue;
      }

      var _result = api.removePreserve(ctx, input[prop], options);
      var container = _getContextValue(ctx, prop, '@container') || [];
      if (options.compactArrays && _isArray(_result) && _result.length === 1 && container.length === 0 && prop !== graphAlias) {
        _result = _result[0];
      }
      input[prop] = _result;
    }
  }
  return input;
};

/**
 * Picks the preferred compaction term from the given inverse context entry.
 *
 * @param activeCtx the active context.
 * @param iri the IRI to pick the term for.
 * @param value the value to pick the term for.
 * @param containers the preferred containers.
 * @param typeOrLanguage either '@type' or '@language'.
 * @param typeOrLanguageValue the preferred value for '@type' or '@language'.
 *
 * @return the preferred term.
 */
function _selectTerm(activeCtx, iri, value, containers, typeOrLanguage, typeOrLanguageValue) {
  if (typeOrLanguageValue === null) {
    typeOrLanguageValue = '@null';
  }

  // preferences for the value of @type or @language
  var prefs = [];

  // determine prefs for @id based on whether or not value compacts to a term
  if ((typeOrLanguageValue === '@id' || typeOrLanguageValue === '@reverse') && _isSubjectReference(value)) {
    // prefer @reverse first
    if (typeOrLanguageValue === '@reverse') {
      prefs.push('@reverse');
    }
    // try to compact value to a term
    var term = api.compactIri({ activeCtx: activeCtx, iri: value['@id'], relativeTo: { vocab: true } });
    if (activeCtx.mappings.has(term) && activeCtx.mappings.get(term) && activeCtx.mappings.get(term)['@id'] === value['@id']) {
      // prefer @vocab
      prefs.push.apply(prefs, ['@vocab', '@id']);
    } else {
      // prefer @id
      prefs.push.apply(prefs, ['@id', '@vocab']);
    }
  } else {
    prefs.push(typeOrLanguageValue);
  }
  prefs.push('@none');

  var containerMap = activeCtx.inverse[iri];
  for (var ci = 0; ci < containers.length; ++ci) {
    // if container not available in the map, continue
    var container = containers[ci];
    if (!(container in containerMap)) {
      continue;
    }

    var typeOrLanguageValueMap = containerMap[container][typeOrLanguage];
    for (var pi = 0; pi < prefs.length; ++pi) {
      // if type/language option not available in the map, continue
      var pref = prefs[pi];
      if (!(pref in typeOrLanguageValueMap)) {
        continue;
      }

      // select term
      return typeOrLanguageValueMap[pref];
    }
  }

  return null;
}

/**
 * The value of `@nest` in the term definition must either be `@nest`, or a term
 * which resolves to `@nest`.
 *
 * @param activeCtx the active context.
 * @param nestProperty a term in the active context or `@nest`.
 * @param {Object} [options] - processing options.
 */
function _checkNestProperty(activeCtx, nestProperty, options) {
  if (_expandIri(activeCtx, nestProperty, { vocab: true }, options) !== '@nest') {
    throw new JsonLdError('JSON-LD compact error; nested property must have an @nest value ' + 'resolving to @nest.', 'jsonld.SyntaxError', { code: 'invalid @nest value' });
  }
}