/*
 * Copyright (c) 2017 Digital Bazaar, Inc. All rights reserved.
 */
'use strict';

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var JsonLdError = require('./JsonLdError');

var _require = require('./types'),
    _isArray = _require.isArray,
    _isObject = _require.isObject,
    _isEmptyObject = _require.isEmptyObject,
    _isString = _require.isString,
    _isUndefined = _require.isUndefined;

var _require2 = require('./graphTypes'),
    _isList = _require2.isList,
    _isValue = _require2.isValue,
    _isGraph = _require2.isGraph;

var _require3 = require('./context'),
    _expandIri = _require3.expandIri,
    _getContextValue = _require3.getContextValue,
    _isKeyword = _require3.isKeyword,
    _processContext = _require3.process;

var _require4 = require('./url'),
    _isAbsoluteIri = _require4.isAbsolute;

var _require5 = require('./util'),
    _addValue = _require5.addValue,
    _asArray = _require5.asArray,
    _getValues = _require5.getValues,
    _validateTypeValue = _require5.validateTypeValue;

var api = {};
module.exports = api;

/**
 * Recursively expands an element using the given context. Any context in
 * the element will be removed. All context URLs must have been retrieved
 * before calling this method.
 *
 * @param activeCtx the context to use.
 * @param activeProperty the property for the element, null for none.
 * @param element the element to expand.
 * @param options the expansion options.
 * @param insideList true if the element is a list, false if not.
 * @param insideIndex true if the element is inside an index container,
 *          false if not.
 * @param typeScopedContext an optional type-scoped active context for
 *          expanding values of nodes that were expressed according to
 *          a type-scoped context.
 * @param expansionMap(info) a function that can be used to custom map
 *          unmappable values (or to throw an error when they are detected);
 *          if this function returns `undefined` then the default behavior
 *          will be used.
 *
 * @return a Promise that resolves to the expanded value.
 */
api.expand = function (_ref) {
  var activeCtx = _ref.activeCtx,
      _ref$activeProperty = _ref.activeProperty,
      activeProperty = _ref$activeProperty === undefined ? null : _ref$activeProperty,
      element = _ref.element,
      _ref$options = _ref.options,
      options = _ref$options === undefined ? {} : _ref$options,
      _ref$insideList = _ref.insideList,
      insideList = _ref$insideList === undefined ? false : _ref$insideList,
      _ref$insideIndex = _ref.insideIndex,
      insideIndex = _ref$insideIndex === undefined ? false : _ref$insideIndex,
      _ref$typeScopedContex = _ref.typeScopedContext,
      typeScopedContext = _ref$typeScopedContex === undefined ? null : _ref$typeScopedContex,
      _ref$expansionMap = _ref.expansionMap,
      expansionMap = _ref$expansionMap === undefined ? function () {
    return undefined;
  } : _ref$expansionMap;

  // nothing to expand
  if (element === null || element === undefined) {
    return null;
  }

  // disable framing if activeProperty is @default
  if (activeProperty === '@default') {
    options = Object.assign({}, options, { isFrame: false });
  }

  if (!_isArray(element) && !_isObject(element)) {
    // drop free-floating scalars that are not in lists unless custom mapped
    if (!insideList && (activeProperty === null || _expandIri(activeCtx, activeProperty, { vocab: true }, options) === '@graph')) {
      // TODO: use `await` to support async
      var mapped = expansionMap({
        unmappedValue: element,
        activeCtx: activeCtx,
        activeProperty: activeProperty,
        options: options,
        insideList: insideList
      });
      if (mapped === undefined) {
        return null;
      }
      return mapped;
    }

    // expand element according to value expansion rules
    return _expandValue({ activeCtx: activeCtx, activeProperty: activeProperty, value: element, options: options });
  }

  // recursively expand array
  if (_isArray(element)) {
    var _rval = [];
    var container = _getContextValue(activeCtx, activeProperty, '@container') || [];
    insideList = insideList || container.includes('@list');
    for (var i = 0; i < element.length; ++i) {
      // expand element
      var e = api.expand({
        activeCtx: activeCtx,
        activeProperty: activeProperty,
        element: element[i],
        options: options,
        expansionMap: expansionMap,
        insideIndex: insideIndex,
        typeScopedContext: typeScopedContext
      });
      if (insideList && (_isArray(e) || _isList(e))) {
        // lists of lists are illegal
        throw new JsonLdError('Invalid JSON-LD syntax; lists of lists are not permitted.', 'jsonld.SyntaxError', { code: 'list of lists' });
      }

      if (e === null) {
        // TODO: add `await` for async support
        e = expansionMap({
          unmappedValue: element[i],
          activeCtx: activeCtx,
          activeProperty: activeProperty,
          parent: element,
          index: i,
          options: options,
          expandedParent: _rval,
          insideList: insideList
        });
        if (e === undefined) {
          continue;
        }
      }

      if (_isArray(e)) {
        _rval = _rval.concat(e);
      } else {
        _rval.push(e);
      }
    }
    return _rval;
  }

  // recursively expand object:

  // first, expand the active property
  var expandedActiveProperty = _expandIri(activeCtx, activeProperty, { vocab: true }, options);

  // second, determine if any type-scoped context should be reverted; it
  // should only be reverted when the following are all true:
  // 1. `element` is not a value or subject reference
  // 2. `insideIndex` is false
  typeScopedContext = typeScopedContext || (activeCtx.previousContext ? activeCtx : null);
  var keys = Object.keys(element).sort();
  var mustRevert = !insideIndex;
  if (mustRevert && typeScopedContext && keys.length <= 2 && !keys.includes('@context')) {
    var _iteratorNormalCompletion = true;
    var _didIteratorError = false;
    var _iteratorError = undefined;

    try {
      for (var _iterator = keys[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
        var key = _step.value;

        var expandedProperty = _expandIri(typeScopedContext, key, { vocab: true }, options);
        if (expandedProperty === '@value') {
          // value found, ensure type-scoped context is used to expand it
          mustRevert = false;
          activeCtx = typeScopedContext;
          break;
        }
        if (expandedProperty === '@id' && keys.length === 1) {
          // subject reference found, do not revert
          mustRevert = false;
          break;
        }
      }
    } catch (err) {
      _didIteratorError = true;
      _iteratorError = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion && _iterator.return) {
          _iterator.return();
        }
      } finally {
        if (_didIteratorError) {
          throw _iteratorError;
        }
      }
    }
  }

  if (mustRevert) {
    // revert type scoped context
    activeCtx = activeCtx.revertTypeScopedContext();
  }

  // if element has a context, process it
  if ('@context' in element) {
    activeCtx = _processContext({ activeCtx: activeCtx, localCtx: element['@context'], options: options });
  }

  // look for scoped contexts on `@type`
  var _iteratorNormalCompletion2 = true;
  var _didIteratorError2 = false;
  var _iteratorError2 = undefined;

  try {
    for (var _iterator2 = keys[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
      var _key = _step2.value;

      var _expandedProperty = _expandIri(activeCtx, _key, { vocab: true }, options);
      if (_expandedProperty === '@type') {
        // set scoped contexts from @type
        // avoid sorting if possible
        var value = element[_key];
        var _types = Array.isArray(value) ? value.length > 1 ? value.slice().sort() : value : [value];
        var _iteratorNormalCompletion3 = true;
        var _didIteratorError3 = false;
        var _iteratorError3 = undefined;

        try {
          for (var _iterator3 = _types[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
            var type = _step3.value;

            var ctx = _getContextValue(activeCtx.previousContext || activeCtx, type, '@context');
            if (!_isUndefined(ctx)) {
              activeCtx = _processContext({
                activeCtx: activeCtx,
                localCtx: ctx,
                options: options,
                isTypeScopedContext: true
              });
            }
          }
        } catch (err) {
          _didIteratorError3 = true;
          _iteratorError3 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion3 && _iterator3.return) {
              _iterator3.return();
            }
          } finally {
            if (_didIteratorError3) {
              throw _iteratorError3;
            }
          }
        }
      }
    }

    // process each key and value in element, ignoring @nest content
  } catch (err) {
    _didIteratorError2 = true;
    _iteratorError2 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion2 && _iterator2.return) {
        _iterator2.return();
      }
    } finally {
      if (_didIteratorError2) {
        throw _iteratorError2;
      }
    }
  }

  var rval = {};
  _expandObject({
    activeCtx: activeCtx,
    activeProperty: activeProperty,
    expandedActiveProperty: expandedActiveProperty,
    element: element,
    expandedParent: rval,
    options: options,
    insideList: insideList,
    typeScopedContext: typeScopedContext,
    expansionMap: expansionMap });

  // get property count on expanded output
  keys = Object.keys(rval);
  var count = keys.length;

  if ('@value' in rval) {
    // @value must only have @language or @type
    if ('@type' in rval && '@language' in rval) {
      throw new JsonLdError('Invalid JSON-LD syntax; an element containing "@value" may not ' + 'contain both "@type" and "@language".', 'jsonld.SyntaxError', { code: 'invalid value object', element: rval });
    }
    var validCount = count - 1;
    if ('@type' in rval) {
      validCount -= 1;
    }
    if ('@index' in rval) {
      validCount -= 1;
    }
    if ('@language' in rval) {
      validCount -= 1;
    }
    if (validCount !== 0) {
      throw new JsonLdError('Invalid JSON-LD syntax; an element containing "@value" may only ' + 'have an "@index" property and at most one other property ' + 'which can be "@type" or "@language".', 'jsonld.SyntaxError', { code: 'invalid value object', element: rval });
    }
    var values = rval['@value'] === null ? [] : _asArray(rval['@value']);
    var types = _getValues(rval, '@type');

    // drop null @values unless custom mapped
    if (values.length === 0) {
      // TODO: use `await` to support async
      var _mapped = expansionMap({
        unmappedValue: rval,
        activeCtx: activeCtx,
        activeProperty: activeProperty,
        element: element,
        options: options,
        insideList: insideList
      });
      if (_mapped !== undefined) {
        rval = _mapped;
      } else {
        rval = null;
      }
    } else if (!values.every(function (v) {
      return _isString(v) || _isEmptyObject(v);
    }) && '@language' in rval) {
      // if @language is present, @value must be a string
      throw new JsonLdError('Invalid JSON-LD syntax; only strings may be language-tagged.', 'jsonld.SyntaxError', { code: 'invalid language-tagged value', element: rval });
    } else if (!types.every(function (t) {
      return _isAbsoluteIri(t) && !(_isString(t) && t.indexOf('_:') === 0) || _isEmptyObject(t);
    })) {
      throw new JsonLdError('Invalid JSON-LD syntax; an element containing "@value" and "@type" ' + 'must have an absolute IRI for the value of "@type".', 'jsonld.SyntaxError', { code: 'invalid typed value', element: rval });
    }
  } else if ('@type' in rval && !_isArray(rval['@type'])) {
    // convert @type to an array
    rval['@type'] = [rval['@type']];
  } else if ('@set' in rval || '@list' in rval) {
    // handle @set and @list
    if (count > 1 && !(count === 2 && '@index' in rval)) {
      throw new JsonLdError('Invalid JSON-LD syntax; if an element has the property "@set" ' + 'or "@list", then it can have at most one other property that is ' + '"@index".', 'jsonld.SyntaxError', { code: 'invalid set or list object', element: rval });
    }
    // optimize away @set
    if ('@set' in rval) {
      rval = rval['@set'];
      keys = Object.keys(rval);
      count = keys.length;
    }
  } else if (count === 1 && '@language' in rval) {
    // drop objects with only @language unless custom mapped
    // TODO: use `await` to support async
    var _mapped2 = expansionMap(rval, {
      unmappedValue: rval,
      activeCtx: activeCtx,
      activeProperty: activeProperty,
      element: element,
      options: options,
      insideList: insideList
    });
    if (_mapped2 !== undefined) {
      rval = _mapped2;
    } else {
      rval = null;
    }
  }

  // drop certain top-level objects that do not occur in lists, unless custom
  // mapped
  if (_isObject(rval) && !options.keepFreeFloatingNodes && !insideList && (activeProperty === null || expandedActiveProperty === '@graph')) {
    // drop empty object, top-level @value/@list, or object with only @id
    if (count === 0 || '@value' in rval || '@list' in rval || count === 1 && '@id' in rval) {
      // TODO: use `await` to support async
      var _mapped3 = expansionMap({
        unmappedValue: rval,
        activeCtx: activeCtx,
        activeProperty: activeProperty,
        element: element,
        options: options,
        insideList: insideList
      });
      if (_mapped3 !== undefined) {
        rval = _mapped3;
      } else {
        rval = null;
      }
    }
  }

  return rval;
};

/**
 * Expand each key and value of element adding to result
 *
 * @param activeCtx the context to use.
 * @param activeProperty the property for the element.
 * @param expandedActiveProperty the expansion of activeProperty
 * @param element the element to expand.
 * @param expandedParent the expanded result into which to add values.
 * @param options the expansion options.
 * @param insideList true if the element is a list, false if not.
 * @param expansionMap(info) a function that can be used to custom map
 *          unmappable values (or to throw an error when they are detected);
 *          if this function returns `undefined` then the default behavior
 *          will be used.
 */
function _expandObject(_ref2) {
  var activeCtx = _ref2.activeCtx,
      activeProperty = _ref2.activeProperty,
      expandedActiveProperty = _ref2.expandedActiveProperty,
      element = _ref2.element,
      expandedParent = _ref2.expandedParent,
      _ref2$options = _ref2.options,
      options = _ref2$options === undefined ? {} : _ref2$options,
      insideList = _ref2.insideList,
      expansionMap = _ref2.expansionMap;

  var keys = Object.keys(element).sort();
  var nests = [];
  var _iteratorNormalCompletion4 = true;
  var _didIteratorError4 = false;
  var _iteratorError4 = undefined;

  try {
    for (var _iterator4 = keys[Symbol.iterator](), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
      var key = _step4.value;

      var value = element[key];
      var expandedValue = void 0;

      // skip @context
      if (key === '@context') {
        continue;
      }

      // expand property
      var expandedProperty = _expandIri(activeCtx, key, { vocab: true }, options);

      // drop non-absolute IRI keys that aren't keywords unless custom mapped
      if (expandedProperty === null || !(_isAbsoluteIri(expandedProperty) || _isKeyword(expandedProperty))) {
        // TODO: use `await` to support async
        expandedProperty = expansionMap({
          unmappedProperty: key,
          activeCtx: activeCtx,
          activeProperty: activeProperty,
          parent: element,
          options: options,
          insideList: insideList,
          value: value,
          expandedParent: expandedParent
        });
        if (expandedProperty === undefined) {
          continue;
        }
      }

      if (_isKeyword(expandedProperty)) {
        if (expandedActiveProperty === '@reverse') {
          throw new JsonLdError('Invalid JSON-LD syntax; a keyword cannot be used as a @reverse ' + 'property.', 'jsonld.SyntaxError', { code: 'invalid reverse property map', value: value });
        }
        if (expandedProperty in expandedParent) {
          throw new JsonLdError('Invalid JSON-LD syntax; colliding keywords detected.', 'jsonld.SyntaxError', { code: 'colliding keywords', keyword: expandedProperty });
        }
      }

      // syntax error if @id is not a string
      if (expandedProperty === '@id') {
        if (!_isString(value)) {
          if (!options.isFrame) {
            throw new JsonLdError('Invalid JSON-LD syntax; "@id" value must a string.', 'jsonld.SyntaxError', { code: 'invalid @id value', value: value });
          }
          if (_isObject(value)) {
            // empty object is a wildcard
            if (!_isEmptyObject(value)) {
              throw new JsonLdError('Invalid JSON-LD syntax; "@id" value an empty object or array ' + 'of strings, if framing', 'jsonld.SyntaxError', { code: 'invalid @id value', value: value });
            }
          } else if (_isArray(value)) {
            if (!value.every(function (v) {
              return _isString(v);
            })) {
              throw new JsonLdError('Invalid JSON-LD syntax; "@id" value an empty object or array ' + 'of strings, if framing', 'jsonld.SyntaxError', { code: 'invalid @id value', value: value });
            }
          } else {
            throw new JsonLdError('Invalid JSON-LD syntax; "@id" value an empty object or array ' + 'of strings, if framing', 'jsonld.SyntaxError', { code: 'invalid @id value', value: value });
          }
        }

        _addValue(expandedParent, '@id', _asArray(value).map(function (v) {
          return _isString(v) ? _expandIri(activeCtx, v, { base: true }, options) : v;
        }), { propertyIsArray: options.isFrame });
        continue;
      }

      if (expandedProperty === '@type') {
        _validateTypeValue(value);
        _addValue(expandedParent, '@type', _asArray(value).map(function (v) {
          return _isString(v) ? _expandIri(activeCtx.previousContext || activeCtx, v, { base: true, vocab: true }, options) : v;
        }), { propertyIsArray: options.isFrame });
        continue;
      }

      // @graph must be an array or an object
      if (expandedProperty === '@graph' && !(_isObject(value) || _isArray(value))) {
        throw new JsonLdError('Invalid JSON-LD syntax; "@graph" value must not be an ' + 'object or an array.', 'jsonld.SyntaxError', { code: 'invalid @graph value', value: value });
      }

      // @value must not be an object or an array (unless framing)
      if (expandedProperty === '@value') {
        if ((_isObject(value) || _isArray(value)) && !options.isFrame) {
          throw new JsonLdError('Invalid JSON-LD syntax; "@value" value must not be an ' + 'object or an array.', 'jsonld.SyntaxError', { code: 'invalid value object value', value: value });
        }

        _addValue(expandedParent, '@value', value, { propertyIsArray: options.isFrame });
        continue;
      }

      // @language must be a string
      if (expandedProperty === '@language') {
        if (value === null) {
          // drop null @language values, they expand as if they didn't exist
          continue;
        }
        if (!_isString(value) && !options.isFrame) {
          throw new JsonLdError('Invalid JSON-LD syntax; "@language" value must be a string.', 'jsonld.SyntaxError', { code: 'invalid language-tagged string', value: value });
        }
        // ensure language value is lowercase
        value = _asArray(value).map(function (v) {
          return _isString(v) ? v.toLowerCase() : v;
        });

        _addValue(expandedParent, '@language', value, { propertyIsArray: options.isFrame });
        continue;
      }

      // @index must be a string
      if (expandedProperty === '@index') {
        if (!_isString(value)) {
          throw new JsonLdError('Invalid JSON-LD syntax; "@index" value must be a string.', 'jsonld.SyntaxError', { code: 'invalid @index value', value: value });
        }
        _addValue(expandedParent, '@index', value);
        continue;
      }

      // @reverse must be an object
      if (expandedProperty === '@reverse') {
        if (!_isObject(value)) {
          throw new JsonLdError('Invalid JSON-LD syntax; "@reverse" value must be an object.', 'jsonld.SyntaxError', { code: 'invalid @reverse value', value: value });
        }

        expandedValue = api.expand({
          activeCtx: activeCtx,
          activeProperty: '@reverse',
          element: value,
          options: options,
          expansionMap: expansionMap
        });
        // properties double-reversed
        if ('@reverse' in expandedValue) {
          for (var property in expandedValue['@reverse']) {
            _addValue(expandedParent, property, expandedValue['@reverse'][property], { propertyIsArray: true });
          }
        }

        // FIXME: can this be merged with code below to simplify?
        // merge in all reversed properties
        var reverseMap = expandedParent['@reverse'] || null;
        for (var _property in expandedValue) {
          if (_property === '@reverse') {
            continue;
          }
          if (reverseMap === null) {
            reverseMap = expandedParent['@reverse'] = {};
          }
          _addValue(reverseMap, _property, [], { propertyIsArray: true });
          var items = expandedValue[_property];
          for (var ii = 0; ii < items.length; ++ii) {
            var item = items[ii];
            if (_isValue(item) || _isList(item)) {
              throw new JsonLdError('Invalid JSON-LD syntax; "@reverse" value must not be a ' + '@value or an @list.', 'jsonld.SyntaxError', { code: 'invalid reverse property value', value: expandedValue });
            }
            _addValue(reverseMap, _property, item, { propertyIsArray: true });
          }
        }

        continue;
      }

      // nested keys
      if (expandedProperty === '@nest') {
        nests.push(key);
        continue;
      }

      // use potential scoped context for key
      var termCtx = activeCtx;
      var ctx = _getContextValue(activeCtx, key, '@context');
      if (!_isUndefined(ctx)) {
        // Note: spec's `from term` var is named `isPropertyTermScopedContext`
        termCtx = _processContext({
          activeCtx: activeCtx,
          localCtx: ctx,
          isPropertyTermScopedContext: true,
          options: options
        });
      }

      var container = _getContextValue(termCtx, key, '@container') || [];

      if (container.includes('@language') && _isObject(value)) {
        // handle language map container (skip if value is not an object)
        expandedValue = _expandLanguageMap(termCtx, value, options);
      } else if (container.includes('@index') && _isObject(value)) {
        // handle index container (skip if value is not an object)
        var asGraph = container.includes('@graph');
        expandedValue = _expandIndexMap({
          activeCtx: termCtx,
          options: options,
          activeProperty: key,
          value: value,
          expansionMap: expansionMap,
          asGraph: asGraph,
          indexKey: '@index'
        });
      } else if (container.includes('@id') && _isObject(value)) {
        // handle id container (skip if value is not an object)
        var _asGraph = container.includes('@graph');
        expandedValue = _expandIndexMap({
          activeCtx: termCtx,
          options: options,
          activeProperty: key,
          value: value,
          expansionMap: expansionMap,
          asGraph: _asGraph,
          indexKey: '@id'
        });
      } else if (container.includes('@type') && _isObject(value)) {
        // handle type container (skip if value is not an object)
        expandedValue = _expandIndexMap({
          // since container is `@type`, revert type scoped context when expanding
          activeCtx: termCtx.revertTypeScopedContext(),
          options: options,
          activeProperty: key,
          value: value,
          expansionMap: expansionMap,
          asGraph: false,
          indexKey: '@type'
        });
      } else {
        // recurse into @list or @set
        var isList = expandedProperty === '@list';
        if (isList || expandedProperty === '@set') {
          var nextActiveProperty = activeProperty;
          if (isList && expandedActiveProperty === '@graph') {
            nextActiveProperty = null;
          }
          expandedValue = api.expand({
            activeCtx: termCtx,
            activeProperty: nextActiveProperty,
            element: value,
            options: options,
            insideList: isList,
            expansionMap: expansionMap
          });
          if (isList && _isList(expandedValue)) {
            throw new JsonLdError('Invalid JSON-LD syntax; lists of lists are not permitted.', 'jsonld.SyntaxError', { code: 'list of lists' });
          }
        } else {
          // recursively expand value with key as new active property
          expandedValue = api.expand({
            activeCtx: termCtx,
            activeProperty: key,
            element: value,
            options: options,
            insideList: false,
            expansionMap: expansionMap
          });
        }
      }

      // drop null values if property is not @value
      if (expandedValue === null && expandedProperty !== '@value') {
        // TODO: use `await` to support async
        expandedValue = expansionMap({
          unmappedValue: value,
          expandedProperty: expandedProperty,
          activeCtx: termCtx,
          activeProperty: activeProperty,
          parent: element,
          options: options,
          insideList: insideList,
          key: key,
          expandedParent: expandedParent
        });
        if (expandedValue === undefined) {
          continue;
        }
      }

      // convert expanded value to @list if container specifies it
      if (expandedProperty !== '@list' && !_isList(expandedValue) && container.includes('@list')) {
        // ensure expanded value in @list is an array
        expandedValue = { '@list': _asArray(expandedValue) };
      }

      // convert expanded value to @graph if container specifies it
      // and value is not, itself, a graph
      // index cases handled above
      if (container.includes('@graph') && !container.some(function (key) {
        return key === '@id' || key === '@index';
      })) {
        // ensure expanded values are arrays
        expandedValue = _asArray(expandedValue).map(function (v) {
          return _isGraph(v) ? v : { '@graph': _asArray(v) };
        });
      }

      // FIXME: can this be merged with code above to simplify?
      // merge in reverse properties
      if (termCtx.mappings.has(key) && termCtx.mappings.get(key).reverse) {
        var _reverseMap = expandedParent['@reverse'] = expandedParent['@reverse'] || {};
        expandedValue = _asArray(expandedValue);
        for (var _ii = 0; _ii < expandedValue.length; ++_ii) {
          var _item = expandedValue[_ii];
          if (_isValue(_item) || _isList(_item)) {
            throw new JsonLdError('Invalid JSON-LD syntax; "@reverse" value must not be a ' + '@value or an @list.', 'jsonld.SyntaxError', { code: 'invalid reverse property value', value: expandedValue });
          }
          _addValue(_reverseMap, expandedProperty, _item, { propertyIsArray: true });
        }
        continue;
      }

      // add value for property
      // use an array except for certain keywords
      var useArray = !['@index', '@id', '@type', '@value', '@language'].includes(expandedProperty);
      _addValue(expandedParent, expandedProperty, expandedValue, {
        propertyIsArray: useArray
      });
    }

    // expand each nested key
  } catch (err) {
    _didIteratorError4 = true;
    _iteratorError4 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion4 && _iterator4.return) {
        _iterator4.return();
      }
    } finally {
      if (_didIteratorError4) {
        throw _iteratorError4;
      }
    }
  }

  var _iteratorNormalCompletion5 = true;
  var _didIteratorError5 = false;
  var _iteratorError5 = undefined;

  try {
    for (var _iterator5 = nests[Symbol.iterator](), _step5; !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
      var _key2 = _step5.value;

      var nestedValues = _isArray(element[_key2]) ? element[_key2] : [element[_key2]];
      var _iteratorNormalCompletion6 = true;
      var _didIteratorError6 = false;
      var _iteratorError6 = undefined;

      try {
        for (var _iterator6 = nestedValues[Symbol.iterator](), _step6; !(_iteratorNormalCompletion6 = (_step6 = _iterator6.next()).done); _iteratorNormalCompletion6 = true) {
          var nv = _step6.value;

          if (!_isObject(nv) || Object.keys(nv).some(function (k) {
            return _expandIri(activeCtx, k, { vocab: true }, options) === '@value';
          })) {
            throw new JsonLdError('Invalid JSON-LD syntax; nested value must be a node object.', 'jsonld.SyntaxError', { code: 'invalid @nest value', value: nv });
          }
          _expandObject({
            activeCtx: activeCtx,
            activeProperty: activeProperty,
            expandedActiveProperty: expandedActiveProperty,
            element: nv,
            expandedParent: expandedParent,
            options: options,
            insideList: insideList,
            expansionMap: expansionMap });
        }
      } catch (err) {
        _didIteratorError6 = true;
        _iteratorError6 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion6 && _iterator6.return) {
            _iterator6.return();
          }
        } finally {
          if (_didIteratorError6) {
            throw _iteratorError6;
          }
        }
      }
    }
  } catch (err) {
    _didIteratorError5 = true;
    _iteratorError5 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion5 && _iterator5.return) {
        _iterator5.return();
      }
    } finally {
      if (_didIteratorError5) {
        throw _iteratorError5;
      }
    }
  }
}

/**
 * Expands the given value by using the coercion and keyword rules in the
 * given context.
 *
 * @param activeCtx the active context to use.
 * @param activeProperty the active property the value is associated with.
 * @param value the value to expand.
 * @param {Object} [options] - processing options.
 *
 * @return the expanded value.
 */
function _expandValue(_ref3) {
  var activeCtx = _ref3.activeCtx,
      activeProperty = _ref3.activeProperty,
      value = _ref3.value,
      options = _ref3.options;

  // nothing to expand
  if (value === null || value === undefined) {
    return null;
  }

  // special-case expand @id and @type (skips '@id' expansion)
  var expandedProperty = _expandIri(activeCtx, activeProperty, { vocab: true }, options);
  if (expandedProperty === '@id') {
    return _expandIri(activeCtx, value, { base: true }, options);
  } else if (expandedProperty === '@type') {
    return _expandIri(activeCtx, value, { vocab: true, base: true }, options);
  }

  // get type definition from context
  var type = _getContextValue(activeCtx, activeProperty, '@type');

  // do @id expansion (automatic for @graph)
  if ((type === '@id' || expandedProperty === '@graph') && _isString(value)) {
    return { '@id': _expandIri(activeCtx, value, { base: true }, options) };
  }
  // do @id expansion w/vocab
  if (type === '@vocab' && _isString(value)) {
    return {
      '@id': _expandIri(activeCtx, value, { vocab: true, base: true }, options)
    };
  }

  // do not expand keyword values
  if (_isKeyword(expandedProperty)) {
    return value;
  }

  var rval = {};

  if (type && !['@id', '@vocab'].includes(type)) {
    // other type
    rval['@type'] = type;
  } else if (_isString(value)) {
    // check for language tagging for strings
    var language = _getContextValue(activeCtx, activeProperty, '@language');
    if (language !== null) {
      rval['@language'] = language;
    }
  }
  // do conversion of values that aren't basic JSON types to strings
  if (!['boolean', 'number', 'string'].includes(typeof value === 'undefined' ? 'undefined' : _typeof(value))) {
    value = value.toString();
  }
  rval['@value'] = value;

  return rval;
}

/**
 * Expands a language map.
 *
 * @param activeCtx the active context to use.
 * @param languageMap the language map to expand.
 * @param {Object} [options] - processing options.
 *
 * @return the expanded language map.
 */
function _expandLanguageMap(activeCtx, languageMap, options) {
  var rval = [];
  var keys = Object.keys(languageMap).sort();
  var _iteratorNormalCompletion7 = true;
  var _didIteratorError7 = false;
  var _iteratorError7 = undefined;

  try {
    for (var _iterator7 = keys[Symbol.iterator](), _step7; !(_iteratorNormalCompletion7 = (_step7 = _iterator7.next()).done); _iteratorNormalCompletion7 = true) {
      var key = _step7.value;

      var expandedKey = _expandIri(activeCtx, key, { vocab: true }, options);
      var val = languageMap[key];
      if (!_isArray(val)) {
        val = [val];
      }
      var _iteratorNormalCompletion8 = true;
      var _didIteratorError8 = false;
      var _iteratorError8 = undefined;

      try {
        for (var _iterator8 = val[Symbol.iterator](), _step8; !(_iteratorNormalCompletion8 = (_step8 = _iterator8.next()).done); _iteratorNormalCompletion8 = true) {
          var item = _step8.value;

          if (item === null) {
            // null values are allowed (8.5) but ignored (3.1)
            continue;
          }
          if (!_isString(item)) {
            throw new JsonLdError('Invalid JSON-LD syntax; language map values must be strings.', 'jsonld.SyntaxError', { code: 'invalid language map value', languageMap: languageMap });
          }
          var _val = { '@value': item };
          if (expandedKey !== '@none') {
            _val['@language'] = key.toLowerCase();
          }
          rval.push(_val);
        }
      } catch (err) {
        _didIteratorError8 = true;
        _iteratorError8 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion8 && _iterator8.return) {
            _iterator8.return();
          }
        } finally {
          if (_didIteratorError8) {
            throw _iteratorError8;
          }
        }
      }
    }
  } catch (err) {
    _didIteratorError7 = true;
    _iteratorError7 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion7 && _iterator7.return) {
        _iterator7.return();
      }
    } finally {
      if (_didIteratorError7) {
        throw _iteratorError7;
      }
    }
  }

  return rval;
}

function _expandIndexMap(_ref4) {
  var activeCtx = _ref4.activeCtx,
      options = _ref4.options,
      activeProperty = _ref4.activeProperty,
      value = _ref4.value,
      expansionMap = _ref4.expansionMap,
      asGraph = _ref4.asGraph,
      indexKey = _ref4.indexKey;

  var rval = [];
  var keys = Object.keys(value).sort();
  var isTypeIndex = indexKey === '@type';
  var _iteratorNormalCompletion9 = true;
  var _didIteratorError9 = false;
  var _iteratorError9 = undefined;

  try {
    for (var _iterator9 = keys[Symbol.iterator](), _step9; !(_iteratorNormalCompletion9 = (_step9 = _iterator9.next()).done); _iteratorNormalCompletion9 = true) {
      var key = _step9.value;

      // if indexKey is @type, there may be a context defined for it
      if (isTypeIndex) {
        var ctx = _getContextValue(activeCtx, key, '@context');
        if (!_isUndefined(ctx)) {
          activeCtx = _processContext({
            activeCtx: activeCtx,
            localCtx: ctx,
            isTypeScopedContext: true,
            options: options
          });
        }
      }

      var val = value[key];
      if (!_isArray(val)) {
        val = [val];
      }

      // expand for @type, but also for @none
      var expandedKey = _expandIri(activeCtx, key, { vocab: true }, options);
      if (indexKey === '@id') {
        // expand document relative
        key = _expandIri(activeCtx, key, { base: true }, options);
      } else if (isTypeIndex) {
        key = expandedKey;
      }

      val = api.expand({
        activeCtx: activeCtx,
        activeProperty: activeProperty,
        element: val,
        options: options,
        insideList: false,
        insideIndex: true,
        expansionMap: expansionMap
      });
      var _iteratorNormalCompletion10 = true;
      var _didIteratorError10 = false;
      var _iteratorError10 = undefined;

      try {
        for (var _iterator10 = val[Symbol.iterator](), _step10; !(_iteratorNormalCompletion10 = (_step10 = _iterator10.next()).done); _iteratorNormalCompletion10 = true) {
          var item = _step10.value;

          // If this is also a @graph container, turn items into graphs
          if (asGraph && !_isGraph(item)) {
            item = { '@graph': [item] };
          }
          if (indexKey === '@type') {
            if (expandedKey === '@none') {
              // ignore @none
            } else if (item['@type']) {
              item['@type'] = [key].concat(item['@type']);
            } else {
              item['@type'] = [key];
            }
          } else if (expandedKey !== '@none' && !(indexKey in item)) {
            item[indexKey] = key;
          }
          rval.push(item);
        }
      } catch (err) {
        _didIteratorError10 = true;
        _iteratorError10 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion10 && _iterator10.return) {
            _iterator10.return();
          }
        } finally {
          if (_didIteratorError10) {
            throw _iteratorError10;
          }
        }
      }
    }
  } catch (err) {
    _didIteratorError9 = true;
    _iteratorError9 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion9 && _iterator9.return) {
        _iterator9.return();
      }
    } finally {
      if (_didIteratorError9) {
        throw _iteratorError9;
      }
    }
  }

  return rval;
}