'use strict';

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

/**
 * A JavaScript implementation of the JSON-LD API.
 *
 * @author Dave Longley
 *
 * @license BSD 3-Clause License
 * Copyright (c) 2011-2017 Digital Bazaar, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the Digital Bazaar, Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
var canonize = require('rdf-canonize');
var util = require('./util');
var IdentifierIssuer = util.IdentifierIssuer;
var JsonLdError = require('./JsonLdError');
var NQuads = require('./NQuads');
var Rdfa = require('./Rdfa');

var _require = require('./expand'),
    _expand = _require.expand;

var _require2 = require('./flatten'),
    _flatten = _require2.flatten;

var _require3 = require('./fromRdf'),
    _fromRDF = _require3.fromRDF;

var _require4 = require('./toRdf'),
    _toRDF = _require4.toRDF;

var _require5 = require('./frame'),
    _frameMergedOrDefault = _require5.frameMergedOrDefault;

var _require6 = require('./types'),
    _isArray = _require6.isArray,
    _isObject = _require6.isObject,
    _isString = _require6.isString;

var _require7 = require('./graphTypes'),
    _isSubjectReference = _require7.isSubjectReference;

var _require8 = require('./context'),
    _getInitialContext = _require8.getInitialContext,
    _processContext = _require8.process,
    _getAllContexts = _require8.getAllContexts;

var _require9 = require('./compact'),
    _compact = _require9.compact,
    _compactIri = _require9.compactIri,
    _removePreserve = _require9.removePreserve;

var _require10 = require('./nodeMap'),
    _createNodeMap = _require10.createNodeMap,
    _createMergedNodeMap = _require10.createMergedNodeMap,
    _mergeNodeMaps = _require10.mergeNodeMaps;

// determine if in-browser or using node.js


var _nodejs = typeof process !== 'undefined' && process.versions && process.versions.node;
var _browser = !_nodejs && (typeof window !== 'undefined' || typeof self !== 'undefined');

/* eslint-disable indent */
// attaches jsonld API to the given object
var wrapper = function wrapper(jsonld) {
  var _this = this;

  /* Core API */

  /**
   * Performs JSON-LD compaction.
   *
   * @param input the JSON-LD input to compact.
   * @param ctx the context to compact with.
   * @param [options] options to use:
   *          [base] the base IRI to use.
   *          [compactArrays] true to compact arrays to single values when
   *            appropriate, false not to (default: true).
   *          [compactToRelative] true to compact IRIs to be relative to document
   *            base, false to keep absolute (default: true)
   *          [graph] true to always output a top-level graph (default: false).
   *          [expandContext] a context to expand with.
   *          [skipExpansion] true to assume the input is expanded and skip
   *            expansion, false not to, defaults to false.
   *          [documentLoader(url, callback(err, remoteDoc))] the document loader.
   *          [expansionMap(info)] a function that can be used to custom map
   *            unmappable values (or to throw an error when they are detected);
   *            if this function returns `undefined` then the default behavior
   *            will be used.
   *          [framing] true if compaction is occuring during a framing operation.
   *          [compactionMap(info)] a function that can be used to custom map
   *            unmappable values (or to throw an error when they are detected);
   *            if this function returns `undefined` then the default behavior
   *            will be used.
   * @param [callback(err, compacted)] called once the operation completes.
   *
   * @return a Promise that resolves to the compacted output.
   */
  jsonld.compact = util.callbackify(function () {
    var _ref = _asyncToGenerator(regeneratorRuntime.mark(function _callee(input, ctx, options) {
      var expanded,
          activeCtx,
          compacted,
          tmp,
          i,
          hasContext,
          graphAlias,
          graph,
          _graph,
          key,
          _graph2,
          _args = arguments;

      return regeneratorRuntime.wrap(function _callee$(_context) {
        while (1) {
          switch (_context.prev = _context.next) {
            case 0:
              if (!(_args.length < 2)) {
                _context.next = 2;
                break;
              }

              throw new TypeError('Could not compact, too few arguments.');

            case 2:
              if (!(ctx === null)) {
                _context.next = 4;
                break;
              }

              throw new JsonLdError('The compaction context must not be null.', 'jsonld.CompactError', { code: 'invalid local context' });

            case 4:
              if (!(input === null)) {
                _context.next = 6;
                break;
              }

              return _context.abrupt('return', null);

            case 6:

              // set default options
              options = _setDefaults(options, {
                base: _isString(input) ? input : '',
                compactArrays: true,
                compactToRelative: true,
                graph: false,
                skipExpansion: false,
                link: false,
                issuer: new IdentifierIssuer('_:b')
              });
              if (options.link) {
                // force skip expansion when linking, "link" is not part of the public
                // API, it should only be called from framing
                options.skipExpansion = true;
              }
              if (!options.compactToRelative) {
                delete options.base;
              }

              // expand input
              expanded = void 0;

              if (!options.skipExpansion) {
                _context.next = 14;
                break;
              }

              expanded = input;
              _context.next = 17;
              break;

            case 14:
              _context.next = 16;
              return jsonld.expand(input, options);

            case 16:
              expanded = _context.sent;

            case 17:
              _context.next = 19;
              return jsonld.processContext(_getInitialContext(options), ctx, options);

            case 19:
              activeCtx = _context.sent;


              // do compaction
              compacted = _compact({
                activeCtx: activeCtx,
                element: expanded,
                options: options,
                compactionMap: options.compactionMap
              });

              // perform clean up

              if (options.compactArrays && !options.graph && _isArray(compacted)) {
                if (compacted.length === 1) {
                  // simplify to a single item
                  compacted = compacted[0];
                } else if (compacted.length === 0) {
                  // simplify to an empty object
                  compacted = {};
                }
              } else if (options.graph && _isObject(compacted)) {
                // always use array if graph option is on
                compacted = [compacted];
              }

              // follow @context key
              if (_isObject(ctx) && '@context' in ctx) {
                ctx = ctx['@context'];
              }

              // build output context
              ctx = util.clone(ctx);
              if (!_isArray(ctx)) {
                ctx = [ctx];
              }
              // remove empty contexts
              tmp = ctx;

              ctx = [];
              for (i = 0; i < tmp.length; ++i) {
                if (!_isObject(tmp[i]) || Object.keys(tmp[i]).length > 0) {
                  ctx.push(tmp[i]);
                }
              }

              // remove array if only one context
              hasContext = ctx.length > 0;

              if (ctx.length === 1) {
                ctx = ctx[0];
              }

              // add context and/or @graph
              if (_isArray(compacted)) {
                // use '@graph' keyword
                graphAlias = _compactIri({
                  activeCtx: activeCtx, iri: '@graph', relativeTo: { vocab: true }
                });
                graph = compacted;

                compacted = {};
                if (hasContext) {
                  compacted['@context'] = ctx;
                }
                compacted[graphAlias] = graph;
              } else if (_isObject(compacted) && hasContext) {
                // reorder keys so @context is first
                _graph = compacted;

                compacted = { '@context': ctx };
                for (key in _graph) {
                  compacted[key] = _graph[key];
                }
              }

              if (options.framing) {
                // get graph alias
                _graph2 = _compactIri({
                  activeCtx: activeCtx, iri: '@graph', relativeTo: { vocab: true }
                });
                // remove @preserve from results

                options.link = {};
                compacted[_graph2] = _removePreserve(activeCtx, compacted[_graph2], options);
              }

              return _context.abrupt('return', compacted);

            case 33:
            case 'end':
              return _context.stop();
          }
        }
      }, _callee, this);
    }));

    return function (_x, _x2, _x3) {
      return _ref.apply(this, arguments);
    };
  }());

  /**
   * Performs JSON-LD expansion.
   *
   * @param input the JSON-LD input to expand.
   * @param [options] the options to use:
   *          [base] the base IRI to use.
   *          [expandContext] a context to expand with.
   *          [keepFreeFloatingNodes] true to keep free-floating nodes,
   *            false not to, defaults to false.
   *          [documentLoader(url, callback(err, remoteDoc))] the document loader.
   *          [expansionMap(info)] a function that can be used to custom map
   *            unmappable values (or to throw an error when they are detected);
   *            if this function returns `undefined` then the default behavior
   *            will be used.
   * @param [callback(err, expanded)] called once the operation completes.
   *
   * @return a Promise that resolves to the expanded output.
   */
  jsonld.expand = util.callbackify(function () {
    var _ref2 = _asyncToGenerator(regeneratorRuntime.mark(function _callee2(input, options) {
      var toResolve,
          contextsToProcess,
          expandContext,
          defaultBase,
          remoteDoc,
          activeCtx,
          expanded,
          _args2 = arguments;
      return regeneratorRuntime.wrap(function _callee2$(_context2) {
        while (1) {
          switch (_context2.prev = _context2.next) {
            case 0:
              if (!(_args2.length < 1)) {
                _context2.next = 2;
                break;
              }

              throw new TypeError('Could not expand, too few arguments.');

            case 2:

              // set default options
              options = _setDefaults(options, {
                keepFreeFloatingNodes: false
              });
              if (options.expansionMap === false) {
                options.expansionMap = undefined;
              }

              // build set of objects that may have @contexts to resolve
              toResolve = {};

              // build set of contexts to process prior to expansion

              contextsToProcess = [];

              // if an `expandContext` has been given ensure it gets resolved

              if ('expandContext' in options) {
                expandContext = util.clone(options.expandContext);

                if (_isObject(expandContext) && '@context' in expandContext) {
                  toResolve.expandContext = expandContext;
                } else {
                  toResolve.expandContext = { '@context': expandContext };
                }
                contextsToProcess.push(toResolve.expandContext);
              }

              // if input is a string, attempt to dereference remote document
              defaultBase = void 0;

              if (_isString(input)) {
                _context2.next = 12;
                break;
              }

              // input is not a URL, do not need to retrieve it first
              toResolve.input = util.clone(input);
              _context2.next = 18;
              break;

            case 12:
              _context2.next = 14;
              return jsonld.get(input, options);

            case 14:
              remoteDoc = _context2.sent;

              defaultBase = remoteDoc.documentUrl;
              toResolve.input = remoteDoc.document;
              if (remoteDoc.contextUrl) {
                // context included in HTTP link header and must be resolved
                toResolve.remoteContext = { '@context': remoteDoc.contextUrl };
                contextsToProcess.push(toResolve.remoteContext);
              }

            case 18:

              // set default base
              if (!('base' in options)) {
                options.base = defaultBase || '';
              }

              // get all contexts in `toResolve`
              _context2.next = 21;
              return _getAllContexts(toResolve, options);

            case 21:

              // process any additional contexts
              activeCtx = _getInitialContext(options);

              contextsToProcess.forEach(function (localCtx) {
                activeCtx = _processContext({ activeCtx: activeCtx, localCtx: localCtx, options: options });
              });

              // expand resolved input
              expanded = _expand({
                activeCtx: activeCtx,
                element: toResolve.input,
                options: options,
                expansionMap: options.expansionMap
              });

              // optimize away @graph with no other properties

              if (_isObject(expanded) && '@graph' in expanded && Object.keys(expanded).length === 1) {
                expanded = expanded['@graph'];
              } else if (expanded === null) {
                expanded = [];
              }

              // normalize to an array
              if (!_isArray(expanded)) {
                expanded = [expanded];
              }

              return _context2.abrupt('return', expanded);

            case 27:
            case 'end':
              return _context2.stop();
          }
        }
      }, _callee2, this);
    }));

    return function (_x4, _x5) {
      return _ref2.apply(this, arguments);
    };
  }());

  /**
   * Performs JSON-LD flattening.
   *
   * @param input the JSON-LD to flatten.
   * @param ctx the context to use to compact the flattened output, or null.
   * @param [options] the options to use:
   *          [base] the base IRI to use.
   *          [expandContext] a context to expand with.
   *          [documentLoader(url, callback(err, remoteDoc))] the document loader.
   * @param [callback(err, flattened)] called once the operation completes.
   *
   * @return a Promise that resolves to the flattened output.
   */
  jsonld.flatten = util.callbackify(function () {
    var _ref3 = _asyncToGenerator(regeneratorRuntime.mark(function _callee3(input, ctx, options) {
      var expanded,
          flattened,
          compacted,
          _args3 = arguments;
      return regeneratorRuntime.wrap(function _callee3$(_context3) {
        while (1) {
          switch (_context3.prev = _context3.next) {
            case 0:
              if (!(_args3.length < 1)) {
                _context3.next = 2;
                break;
              }

              return _context3.abrupt('return', new TypeError('Could not flatten, too few arguments.'));

            case 2:

              if (typeof ctx === 'function') {
                ctx = null;
              } else {
                ctx = ctx || null;
              }

              // set default options
              options = _setDefaults(options, {
                base: _isString(input) ? input : ''
              });

              // expand input
              _context3.next = 6;
              return jsonld.expand(input, options);

            case 6:
              expanded = _context3.sent;


              // do flattening
              flattened = _flatten(expanded);

              if (!(ctx === null)) {
                _context3.next = 10;
                break;
              }

              return _context3.abrupt('return', flattened);

            case 10:

              // compact result (force @graph option to true, skip expansion)
              options.graph = true;
              options.skipExpansion = true;
              _context3.next = 14;
              return jsonld.compact(flattened, ctx, options);

            case 14:
              compacted = _context3.sent;
              return _context3.abrupt('return', compacted);

            case 16:
            case 'end':
              return _context3.stop();
          }
        }
      }, _callee3, this);
    }));

    return function (_x6, _x7, _x8) {
      return _ref3.apply(this, arguments);
    };
  }());

  /**
   * Performs JSON-LD framing.
   *
   * @param input the JSON-LD input to frame.
   * @param frame the JSON-LD frame to use.
   * @param [options] the framing options.
   *          [base] the base IRI to use.
   *          [expandContext] a context to expand with.
   *          [embed] default @embed flag: '@last', '@always', '@never', '@link'
   *            (default: '@last').
   *          [explicit] default @explicit flag (default: false).
   *          [requireAll] default @requireAll flag (default: true).
   *          [omitDefault] default @omitDefault flag (default: false).
   *          [documentLoader(url, callback(err, remoteDoc))] the document loader.
   * @param [callback(err, framed)] called once the operation completes.
   *
   * @return a Promise that resolves to the framed output.
   */
  jsonld.frame = util.callbackify(function () {
    var _ref4 = _asyncToGenerator(regeneratorRuntime.mark(function _callee4(input, frame, options) {
      var remoteDoc,
          ctx,
          frameContext,
          expanded,
          opts,
          expandedFrame,
          framed,
          compacted,
          _args4 = arguments;
      return regeneratorRuntime.wrap(function _callee4$(_context4) {
        while (1) {
          switch (_context4.prev = _context4.next) {
            case 0:
              if (!(_args4.length < 2)) {
                _context4.next = 2;
                break;
              }

              throw new TypeError('Could not frame, too few arguments.');

            case 2:

              // set default options
              options = _setDefaults(options, {
                base: _isString(input) ? input : '',
                embed: '@last',
                explicit: false,
                requireAll: true,
                omitDefault: false,
                pruneBlankNodeIdentifiers: true,
                bnodesToClear: []
              });

              // if frame is a string, attempt to dereference remote document

              if (!_isString(frame)) {
                _context4.next = 9;
                break;
              }

              _context4.next = 6;
              return jsonld.get(frame, options);

            case 6:
              remoteDoc = _context4.sent;

              frame = remoteDoc.document;

              if (remoteDoc.contextUrl) {
                // inject link header @context into frame
                ctx = frame['@context'];

                if (!ctx) {
                  ctx = remoteDoc.contextUrl;
                } else if (_isArray(ctx)) {
                  ctx.push(remoteDoc.contextUrl);
                } else {
                  ctx = [ctx, remoteDoc.contextUrl];
                }
                frame['@context'] = ctx;
              }

            case 9:
              frameContext = frame ? frame['@context'] || {} : {};

              // expand input

              _context4.next = 12;
              return jsonld.expand(input, options);

            case 12:
              expanded = _context4.sent;


              // expand frame
              opts = util.clone(options);

              opts.isFrame = true;
              opts.keepFreeFloatingNodes = true;
              _context4.next = 18;
              return jsonld.expand(frame, opts);

            case 18:
              expandedFrame = _context4.sent;


              // if the unexpanded frame includes a key expanding to @graph, frame the
              // default graph, otherwise, the merged graph
              framed = void 0;
              // FIXME should look for aliases of @graph

              opts.merged = !('@graph' in frame);
              // do framing
              framed = _frameMergedOrDefault(expanded, expandedFrame, opts);

              // compact result (force @graph option to true, skip expansion,
              // check for linked embeds)
              opts.graph = true;
              opts.skipExpansion = true;
              opts.link = {};
              opts.framing = true;
              _context4.next = 28;
              return jsonld.compact(framed, frameContext, opts);

            case 28:
              compacted = _context4.sent;
              return _context4.abrupt('return', compacted);

            case 30:
            case 'end':
              return _context4.stop();
          }
        }
      }, _callee4, this);
    }));

    return function (_x9, _x10, _x11) {
      return _ref4.apply(this, arguments);
    };
  }());

  /**
   * **Experimental**
   *
   * Links a JSON-LD document's nodes in memory.
   *
   * @param input the JSON-LD document to link.
   * @param [ctx] the JSON-LD context to apply.
   * @param [options] the options to use:
   *          [base] the base IRI to use.
   *          [expandContext] a context to expand with.
   *          [documentLoader(url, callback(err, remoteDoc))] the document loader.
   * @param [callback(err, linked)] called once the operation completes.
   *
   * @return a Promise that resolves to the linked output.
   */
  jsonld.link = util.callbackify(function () {
    var _ref5 = _asyncToGenerator(regeneratorRuntime.mark(function _callee5(input, ctx, options) {
      var frame;
      return regeneratorRuntime.wrap(function _callee5$(_context5) {
        while (1) {
          switch (_context5.prev = _context5.next) {
            case 0:
              // API matches running frame with a wildcard frame and embed: '@link'
              // get arguments
              frame = {};

              if (ctx) {
                frame['@context'] = ctx;
              }
              frame['@embed'] = '@link';
              return _context5.abrupt('return', jsonld.frame(input, frame, options));

            case 4:
            case 'end':
              return _context5.stop();
          }
        }
      }, _callee5, this);
    }));

    return function (_x12, _x13, _x14) {
      return _ref5.apply(this, arguments);
    };
  }());

  /**
   * Performs RDF dataset normalization on the given input. The input is JSON-LD
   * unless the 'inputFormat' option is used. The output is an RDF dataset
   * unless the 'format' option is used.
   *
   * @param input the input to normalize as JSON-LD or as a format specified by
   *          the 'inputFormat' option.
   * @param [options] the options to use:
   *          [algorithm] the normalization algorithm to use, `URDNA2015` or
   *            `URGNA2012` (default: `URDNA2015`).
   *          [base] the base IRI to use.
   *          [expandContext] a context to expand with.
   *          [skipExpansion] true to assume the input is expanded and skip
   *            expansion, false not to, defaults to false.
   *          [inputFormat] the format if input is not JSON-LD:
   *            'application/n-quads' for N-Quads.
   *          [format] the format if output is a string:
   *            'application/n-quads' for N-Quads.
   *          [documentLoader(url, callback(err, remoteDoc))] the document loader.
   *          [useNative] true to use a native canonize algorithm
   * @param [callback(err, normalized)] called once the operation completes.
   *
   * @return a Promise that resolves to the normalized output.
   */
  jsonld.normalize = jsonld.canonize = util.callbackify(function () {
    var _ref6 = _asyncToGenerator(regeneratorRuntime.mark(function _callee6(input, options) {
      var parsedInput,
          opts,
          dataset,
          _args6 = arguments;
      return regeneratorRuntime.wrap(function _callee6$(_context6) {
        while (1) {
          switch (_context6.prev = _context6.next) {
            case 0:
              if (!(_args6.length < 1)) {
                _context6.next = 2;
                break;
              }

              throw new TypeError('Could not canonize, too few arguments.');

            case 2:

              // set default options
              options = _setDefaults(options, {
                base: _isString(input) ? input : '',
                algorithm: 'URDNA2015',
                skipExpansion: false
              });

              if (!('inputFormat' in options)) {
                _context6.next = 8;
                break;
              }

              if (!(options.inputFormat !== 'application/n-quads' && options.inputFormat !== 'application/nquads')) {
                _context6.next = 6;
                break;
              }

              throw new JsonLdError('Unknown canonicalization input format.', 'jsonld.CanonizeError');

            case 6:
              // TODO: `await` for async parsers
              parsedInput = NQuads.parse(input);

              // do canonicalization

              return _context6.abrupt('return', canonize.canonize(parsedInput, options));

            case 8:

              // convert to RDF dataset then do normalization
              opts = util.clone(options);

              delete opts.format;
              opts.produceGeneralizedRdf = false;
              _context6.next = 13;
              return jsonld.toRDF(input, opts);

            case 13:
              dataset = _context6.sent;
              return _context6.abrupt('return', canonize.canonize(dataset, options));

            case 15:
            case 'end':
              return _context6.stop();
          }
        }
      }, _callee6, this);
    }));

    return function (_x15, _x16) {
      return _ref6.apply(this, arguments);
    };
  }());

  /**
   * Converts an RDF dataset to JSON-LD.
   *
   * @param dataset a serialized string of RDF in a format specified by the
   *          format option or an RDF dataset to convert.
   * @param [options] the options to use:
   *          [format] the format if dataset param must first be parsed:
   *            'application/n-quads' for N-Quads (default).
   *          [rdfParser] a custom RDF-parser to use to parse the dataset.
   *          [useRdfType] true to use rdf:type, false to use @type
   *            (default: false).
   *          [useNativeTypes] true to convert XSD types into native types
   *            (boolean, integer, double), false not to (default: false).
   * @param [callback(err, output)] called once the operation completes.
   *
   * @return a Promise that resolves to the JSON-LD document.
   */
  jsonld.fromRDF = util.callbackify(function () {
    var _ref7 = _asyncToGenerator(regeneratorRuntime.mark(function _callee7(dataset, options) {
      var _options,
          format,
          rdfParser,
          parsedDataset,
          _args7 = arguments;

      return regeneratorRuntime.wrap(function _callee7$(_context7) {
        while (1) {
          switch (_context7.prev = _context7.next) {
            case 0:
              if (!(_args7.length < 1)) {
                _context7.next = 2;
                break;
              }

              throw new TypeError('Could not convert from RDF, too few arguments.');

            case 2:

              // set default options
              options = _setDefaults(options, {
                format: _isString(dataset) ? 'application/n-quads' : undefined
              });

              _options = options, format = _options.format, rdfParser = _options.rdfParser;

              // handle special format

              if (!format) {
                _context7.next = 10;
                break;
              }

              // check supported formats
              rdfParser = rdfParser || _rdfParsers[format];

              if (rdfParser) {
                _context7.next = 8;
                break;
              }

              throw new JsonLdError('Unknown input format.', 'jsonld.UnknownFormat', { format: format });

            case 8:
              _context7.next = 11;
              break;

            case 10:
              // no-op parser, assume dataset already parsed
              rdfParser = function rdfParser() {
                return dataset;
              };

            case 11:

              // TODO: call `normalizeAsyncFn` on parser fn

              // rdfParser can be callback, promise-based, or synchronous
              parsedDataset = void 0;

              if (rdfParser.length > 1) {
                // convert callback-based rdf parser to promise-based
                parsedDataset = new Promise(function (resolve, reject) {
                  rdfParser(dataset, function (err, dataset) {
                    if (err) {
                      reject(err);
                    } else {
                      resolve(dataset);
                    }
                  });
                });
              } else {
                parsedDataset = Promise.resolve(rdfParser(dataset));
              }

              _context7.next = 15;
              return parsedDataset;

            case 15:
              parsedDataset = _context7.sent;


              // back-compat with old parsers that produced legacy dataset format
              if (!Array.isArray(parsedDataset)) {
                parsedDataset = NQuads.legacyDatasetToQuads(parsedDataset);
              }

              return _context7.abrupt('return', _fromRDF(parsedDataset, options));

            case 18:
            case 'end':
              return _context7.stop();
          }
        }
      }, _callee7, this);
    }));

    return function (_x17, _x18) {
      return _ref7.apply(this, arguments);
    };
  }());

  /**
   * Outputs the RDF dataset found in the given JSON-LD object.
   *
   * @param input the JSON-LD input.
   * @param [options] the options to use:
   *          [base] the base IRI to use.
   *          [expandContext] a context to expand with.
   *          [skipExpansion] true to assume the input is expanded and skip
   *            expansion, false not to, defaults to false.
   *          [format] the format to use to output a string:
   *            'application/n-quads' for N-Quads.
   *          [produceGeneralizedRdf] true to output generalized RDF, false
   *            to produce only standard RDF (default: false).
   *          [documentLoader(url, callback(err, remoteDoc))] the document loader.
   * @param [callback(err, dataset)] called once the operation completes.
   *
   * @return a Promise that resolves to the RDF dataset.
   */
  jsonld.toRDF = util.callbackify(function () {
    var _ref8 = _asyncToGenerator(regeneratorRuntime.mark(function _callee8(input, options) {
      var expanded,
          dataset,
          _args8 = arguments;
      return regeneratorRuntime.wrap(function _callee8$(_context8) {
        while (1) {
          switch (_context8.prev = _context8.next) {
            case 0:
              if (!(_args8.length < 1)) {
                _context8.next = 2;
                break;
              }

              throw new TypeError('Could not convert to RDF, too few arguments.');

            case 2:

              // set default options
              options = _setDefaults(options, {
                base: _isString(input) ? input : '',
                skipExpansion: false
              });

              // TODO: support toRDF custom map?
              expanded = void 0;

              if (!options.skipExpansion) {
                _context8.next = 8;
                break;
              }

              expanded = input;
              _context8.next = 11;
              break;

            case 8:
              _context8.next = 10;
              return jsonld.expand(input, options);

            case 10:
              expanded = _context8.sent;

            case 11:

              // output RDF dataset
              dataset = _toRDF(expanded, options);

              if (!options.format) {
                _context8.next = 18;
                break;
              }

              if (!(options.format === 'application/n-quads' || options.format === 'application/nquads')) {
                _context8.next = 17;
                break;
              }

              _context8.next = 16;
              return NQuads.serialize(dataset);

            case 16:
              return _context8.abrupt('return', _context8.sent);

            case 17:
              throw new JsonLdError('Unknown output format.', 'jsonld.UnknownFormat', { format: options.format });

            case 18:
              return _context8.abrupt('return', dataset);

            case 19:
            case 'end':
              return _context8.stop();
          }
        }
      }, _callee8, this);
    }));

    return function (_x19, _x20) {
      return _ref8.apply(this, arguments);
    };
  }());

  /**
   * **Experimental**
   *
   * Recursively flattens the nodes in the given JSON-LD input into a merged
   * map of node ID => node. All graphs will be merged into the default graph.
   *
   * @param input the JSON-LD input.
   * @param [options] the options to use:
   *          [base] the base IRI to use.
   *          [expandContext] a context to expand with.
   *          [issuer] a jsonld.IdentifierIssuer to use to label blank nodes.
   *          [documentLoader(url, callback(err, remoteDoc))] the document loader.
   * @param [callback(err, nodeMap)] called once the operation completes.
   *
   * @return a Promise that resolves to the merged node map.
   */
  jsonld.createNodeMap = util.callbackify(function () {
    var _ref9 = _asyncToGenerator(regeneratorRuntime.mark(function _callee9(input, options) {
      var expanded,
          _args9 = arguments;
      return regeneratorRuntime.wrap(function _callee9$(_context9) {
        while (1) {
          switch (_context9.prev = _context9.next) {
            case 0:
              if (!(_args9.length < 1)) {
                _context9.next = 2;
                break;
              }

              throw new TypeError('Could not create node map, too few arguments.');

            case 2:

              // set default options
              options = _setDefaults(options, {
                base: _isString(input) ? input : ''
              });

              // expand input
              _context9.next = 5;
              return jsonld.expand(input, options);

            case 5:
              expanded = _context9.sent;
              return _context9.abrupt('return', _createMergedNodeMap(expanded, options));

            case 7:
            case 'end':
              return _context9.stop();
          }
        }
      }, _callee9, this);
    }));

    return function (_x21, _x22) {
      return _ref9.apply(this, arguments);
    };
  }());

  /**
   * **Experimental**
   *
   * Merges two or more JSON-LD documents into a single flattened document.
   *
   * @param docs the JSON-LD documents to merge together.
   * @param ctx the context to use to compact the merged result, or null.
   * @param [options] the options to use:
   *          [base] the base IRI to use.
   *          [expandContext] a context to expand with.
   *          [issuer] a jsonld.IdentifierIssuer to use to label blank nodes.
   *          [mergeNodes] true to merge properties for nodes with the same ID,
   *            false to ignore new properties for nodes with the same ID once
   *            the ID has been defined; note that this may not prevent merging
   *            new properties where a node is in the `object` position
   *            (default: true).
   *          [documentLoader(url, callback(err, remoteDoc))] the document loader.
   * @param [callback(err, merged)] called once the operation completes.
   *
   * @return a Promise that resolves to the merged output.
   */
  jsonld.merge = util.callbackify(function () {
    var _ref10 = _asyncToGenerator(regeneratorRuntime.mark(function _callee10(docs, ctx, options) {
      var expanded,
          mergeNodes,
          issuer,
          graphs,
          i,
          doc,
          _graphs,
          graphName,
          _nodeMap,
          nodeMap,
          key,
          defaultGraph,
          flattened,
          keys,
          ki,
          node,
          compacted,
          _args10 = arguments;

      return regeneratorRuntime.wrap(function _callee10$(_context10) {
        while (1) {
          switch (_context10.prev = _context10.next) {
            case 0:
              if (!(_args10.length < 1)) {
                _context10.next = 2;
                break;
              }

              throw new TypeError('Could not merge, too few arguments.');

            case 2:
              if (_isArray(docs)) {
                _context10.next = 4;
                break;
              }

              throw new TypeError('Could not merge, "docs" must be an array.');

            case 4:

              if (typeof ctx === 'function') {
                ctx = null;
              } else {
                ctx = ctx || null;
              }

              // set default options
              options = _setDefaults(options, {});

              // expand all documents
              _context10.next = 8;
              return Promise.all(docs.map(function (doc) {
                var opts = Object.assign({}, options);
                return jsonld.expand(doc, opts);
              }));

            case 8:
              expanded = _context10.sent;
              mergeNodes = true;

              if ('mergeNodes' in options) {
                mergeNodes = options.mergeNodes;
              }

              issuer = options.issuer || new IdentifierIssuer('_:b');
              graphs = { '@default': {} };
              i = 0;

            case 14:
              if (!(i < expanded.length)) {
                _context10.next = 33;
                break;
              }

              // uniquely relabel blank nodes
              doc = util.relabelBlankNodes(expanded[i], {
                issuer: new IdentifierIssuer('_:b' + i + '-')
              });

              // add nodes to the shared node map graphs if merging nodes, to a
              // separate graph set if not

              _graphs = mergeNodes || i === 0 ? graphs : { '@default': {} };

              _createNodeMap(doc, _graphs, '@default', issuer);

              if (!(_graphs !== graphs)) {
                _context10.next = 30;
                break;
              }

              _context10.t0 = regeneratorRuntime.keys(_graphs);

            case 20:
              if ((_context10.t1 = _context10.t0()).done) {
                _context10.next = 30;
                break;
              }

              graphName = _context10.t1.value;
              _nodeMap = _graphs[graphName];

              if (graphName in graphs) {
                _context10.next = 26;
                break;
              }

              graphs[graphName] = _nodeMap;
              return _context10.abrupt('continue', 20);

            case 26:
              nodeMap = graphs[graphName];

              for (key in _nodeMap) {
                if (!(key in nodeMap)) {
                  nodeMap[key] = _nodeMap[key];
                }
              }
              _context10.next = 20;
              break;

            case 30:
              ++i;
              _context10.next = 14;
              break;

            case 33:

              // add all non-default graphs to default graph
              defaultGraph = _mergeNodeMaps(graphs);

              // produce flattened output

              flattened = [];
              keys = Object.keys(defaultGraph).sort();

              for (ki = 0; ki < keys.length; ++ki) {
                node = defaultGraph[keys[ki]];
                // only add full subjects to top-level

                if (!_isSubjectReference(node)) {
                  flattened.push(node);
                }
              }

              if (!(ctx === null)) {
                _context10.next = 39;
                break;
              }

              return _context10.abrupt('return', flattened);

            case 39:

              // compact result (force @graph option to true, skip expansion)
              options.graph = true;
              options.skipExpansion = true;
              _context10.next = 43;
              return jsonld.compact(flattened, ctx, options);

            case 43:
              compacted = _context10.sent;
              return _context10.abrupt('return', compacted);

            case 45:
            case 'end':
              return _context10.stop();
          }
        }
      }, _callee10, this);
    }));

    return function (_x23, _x24, _x25) {
      return _ref10.apply(this, arguments);
    };
  }());

  /**
   * The default document loader for external documents. If the environment
   * is node.js, a callback-continuation-style document loader is used; otherwise,
   * a promises-style document loader is used.
   *
   * @param url the URL to load.
   * @param callback(err, remoteDoc) called once the operation completes,
   *          if using a non-promises API.
   *
   * @return a promise, if using a promises API.
   */
  Object.defineProperty(jsonld, 'documentLoader', {
    get: function get() {
      return jsonld._documentLoader;
    },
    set: function set(v) {
      return jsonld._documentLoader = util.normalizeDocumentLoader(v);
    }
  });
  // default document loader not implemented
  jsonld.documentLoader = function () {
    var _ref11 = _asyncToGenerator(regeneratorRuntime.mark(function _callee11(url) {
      return regeneratorRuntime.wrap(function _callee11$(_context11) {
        while (1) {
          switch (_context11.prev = _context11.next) {
            case 0:
              throw new JsonLdError('Could not retrieve a JSON-LD document from the URL. URL ' + 'dereferencing not implemented.', 'jsonld.LoadDocumentError', { code: 'loading document failed', url: url });

            case 1:
            case 'end':
              return _context11.stop();
          }
        }
      }, _callee11, _this);
    }));

    return function (_x26) {
      return _ref11.apply(this, arguments);
    };
  }();

  /**
   * Deprecated default document loader. Do not use or override.
   */
  jsonld.loadDocument = util.callbackify(_asyncToGenerator(regeneratorRuntime.mark(function _callee12() {
    var _args12 = arguments;
    return regeneratorRuntime.wrap(function _callee12$(_context12) {
      while (1) {
        switch (_context12.prev = _context12.next) {
          case 0:
            return _context12.abrupt('return', jsonld.documentLoader.apply(null, _args12));

          case 1:
          case 'end':
            return _context12.stop();
        }
      }
    }, _callee12, this);
  })));

  /**
   * Gets a remote JSON-LD document using the default document loader or
   * one given in the passed options.
   *
   * @param url the URL to fetch.
   * @param [options] the options to use:
   *          [documentLoader] the document loader to use.
   * @param [callback(err, remoteDoc)] called once the operation completes.
   *
   * @return a Promise that resolves to the retrieved remote document.
   */
  jsonld.get = util.callbackify(function () {
    var _ref13 = _asyncToGenerator(regeneratorRuntime.mark(function _callee13(url, options) {
      var load, remoteDoc;
      return regeneratorRuntime.wrap(function _callee13$(_context13) {
        while (1) {
          switch (_context13.prev = _context13.next) {
            case 0:
              load = void 0;

              if (typeof options.documentLoader === 'function') {
                load = util.normalizeDocumentLoader(options.documentLoader);
              } else {
                load = jsonld.documentLoader;
              }

              _context13.next = 4;
              return load(url);

            case 4:
              remoteDoc = _context13.sent;
              _context13.prev = 5;

              if (remoteDoc.document) {
                _context13.next = 8;
                break;
              }

              throw new JsonLdError('No remote document found at the given URL.', 'jsonld.NullRemoteDocument');

            case 8:
              if (_isString(remoteDoc.document)) {
                remoteDoc.document = JSON.parse(remoteDoc.document);
              }
              _context13.next = 14;
              break;

            case 11:
              _context13.prev = 11;
              _context13.t0 = _context13['catch'](5);
              throw new JsonLdError('Could not retrieve a JSON-LD document from the URL.', 'jsonld.LoadDocumentError', {
                code: 'loading document failed',
                cause: _context13.t0,
                remoteDoc: remoteDoc
              });

            case 14:
              return _context13.abrupt('return', remoteDoc);

            case 15:
            case 'end':
              return _context13.stop();
          }
        }
      }, _callee13, this, [[5, 11]]);
    }));

    return function (_x27, _x28) {
      return _ref13.apply(this, arguments);
    };
  }());

  /**
   * Processes a local context, resolving any URLs as necessary, and returns a
   * new active context in its callback.
   *
   * @param activeCtx the current active context.
   * @param localCtx the local context to process.
   * @param [options] the options to use:
   *          [documentLoader(url, callback(err, remoteDoc))] the document loader.
   * @param [callback(err, activeCtx)] called once the operation completes.
   *
   * @return a Promise that resolves to the new active context.
   */
  jsonld.processContext = util.callbackify(function () {
    var _ref14 = _asyncToGenerator(regeneratorRuntime.mark(function _callee14(activeCtx, localCtx, options) {
      var ctx;
      return regeneratorRuntime.wrap(function _callee14$(_context14) {
        while (1) {
          switch (_context14.prev = _context14.next) {
            case 0:
              // set default options
              options = _setDefaults(options, {
                base: ''
              });

              // return initial context early for null context

              if (!(localCtx === null)) {
                _context14.next = 3;
                break;
              }

              return _context14.abrupt('return', _getInitialContext(options));

            case 3:

              // get URLs in localCtx
              localCtx = util.clone(localCtx);
              if (!(_isObject(localCtx) && '@context' in localCtx)) {
                localCtx = { '@context': localCtx };
              }
              _context14.next = 7;
              return _getAllContexts(localCtx, options);

            case 7:
              ctx = _context14.sent;
              return _context14.abrupt('return', _processContext({ activeCtx: activeCtx, localCtx: ctx, options: options }));

            case 9:
            case 'end':
              return _context14.stop();
          }
        }
      }, _callee14, this);
    }));

    return function (_x29, _x30, _x31) {
      return _ref14.apply(this, arguments);
    };
  }());

  // backwards compatibility
  jsonld.getContextValue = require('./context').getContextValue;

  /**
   * Document loaders.
   */
  jsonld.documentLoaders = {};
  jsonld.documentLoaders.node = require('./documentLoaders/node');
  jsonld.documentLoaders.xhr = require('./documentLoaders/xhr');

  /**
   * Assigns the default document loader for external document URLs to a built-in
   * default. Supported types currently include: 'xhr' and 'node'.
   *
   * @param type the type to set.
   * @param [params] the parameters required to use the document loader.
   */
  jsonld.useDocumentLoader = function (type) {
    if (!(type in jsonld.documentLoaders)) {
      throw new JsonLdError('Unknown document loader type: "' + type + '"', 'jsonld.UnknownDocumentLoader', { type: type });
    }

    // set document loader
    jsonld.documentLoader = jsonld.documentLoaders[type].apply(jsonld, Array.prototype.slice.call(arguments, 1));
  };

  /** Registered RDF dataset parsers hashed by content-type. */
  var _rdfParsers = {};

  /**
   * Registers an RDF dataset parser by content-type, for use with
   * jsonld.fromRDF. An RDF dataset parser will always be given two parameters,
   * a string of input and a callback. An RDF dataset parser can be synchronous
   * or asynchronous.
   *
   * If the parser function returns undefined or null then it will be assumed to
   * be asynchronous w/a continuation-passing style and the callback parameter
   * given to the parser MUST be invoked.
   *
   * If it returns a Promise, then it will be assumed to be asynchronous, but the
   * callback parameter MUST NOT be invoked. It should instead be ignored.
   *
   * If it returns an RDF dataset, it will be assumed to be synchronous and the
   * callback parameter MUST NOT be invoked. It should instead be ignored.
   *
   * @param contentType the content-type for the parser.
   * @param parser(input, callback(err, dataset)) the parser function (takes a
   *          string as a parameter and either returns null/undefined and uses
   *          the given callback, returns a Promise, or returns an RDF dataset).
   */
  jsonld.registerRDFParser = function (contentType, parser) {
    _rdfParsers[contentType] = parser;
  };

  /**
   * Unregisters an RDF dataset parser by content-type.
   *
   * @param contentType the content-type for the parser.
   */
  jsonld.unregisterRDFParser = function (contentType) {
    delete _rdfParsers[contentType];
  };

  // register the N-Quads RDF parser
  jsonld.registerRDFParser('application/n-quads', NQuads.parse);
  jsonld.registerRDFParser('application/nquads', NQuads.parse);

  // register the RDFa API RDF parser
  jsonld.registerRDFParser('rdfa-api', Rdfa.parse);

  /* URL API */
  jsonld.url = require('./url');

  /* Utility API */
  jsonld.util = util;
  // backwards compatibility
  Object.assign(jsonld, util);

  // reexpose API as jsonld.promises for backwards compatability
  jsonld.promises = jsonld;

  // backwards compatibility
  jsonld.RequestQueue = require('./RequestQueue');

  /* WebIDL API */
  jsonld.JsonLdProcessor = require('./JsonLdProcessor')(jsonld);

  // setup browser global JsonLdProcessor
  if (_browser && typeof global.JsonLdProcessor === 'undefined') {
    Object.defineProperty(global, 'JsonLdProcessor', {
      writable: true,
      enumerable: false,
      configurable: true,
      value: jsonld.JsonLdProcessor
    });
  }

  // set platform-specific defaults/APIs
  if (_nodejs) {
    // use node document loader by default
    jsonld.useDocumentLoader('node');
  } else if (typeof XMLHttpRequest !== 'undefined') {
    // use xhr document loader by default
    jsonld.useDocumentLoader('xhr');
  }

  function _setDefaults(options, _ref15) {
    var _ref15$documentLoader = _ref15.documentLoader,
        documentLoader = _ref15$documentLoader === undefined ? jsonld.documentLoader : _ref15$documentLoader,
        defaults = _ref15.defaults;

    return Object.assign({}, { documentLoader: documentLoader }, defaults, options);
  }

  // end of jsonld API `wrapper` factory
  return jsonld;
};

// external APIs:

// used to generate a new jsonld API instance
var factory = function factory() {
  return wrapper(function () {
    return factory();
  });
};

// wrap the main jsonld API instance
wrapper(factory);
// export API
module.exports = factory;